#!/usr/bin/env python
#  -*- coding: utf-8 -*-
__author__ = "Ringo"

'''
自动扶梯 策略 (难度：初级)
参考: https://www.shinnytech.com/blog/escalator/
注: 该示例策略仅用于功能示范, 实盘时请根据自己的策略/经验进行修改
'''

from tqsdk import TqApi, TqAuth, TargetPosTask
from tqsdk.ta import MA

# 设置合约
SYMBOL = "SHFE.rb2012"
# 设置均线长短周期
MA_SLOW, MA_FAST = 8, 40

api = TqApi(auth=TqAuth("快期账户", "账户密码"))
klines = api.get_kline_serial(SYMBOL, 60 * 60 * 24)
quote = api.get_quote(SYMBOL)
position = api.get_position(SYMBOL)
target_pos = TargetPosTask(api, SYMBOL)


# K线收盘价在这根K线波动范围函数


def kline_range(num):
    kl_range = (klines.iloc[num].close - klines.iloc[num].low) / \
               (klines.iloc[num].high - klines.iloc[num].low)
    return kl_range


# 获取长短均线值
def ma_caculate(klines):
    ma_slow = MA(klines, MA_SLOW).iloc[-1].ma
    ma_fast = MA(klines, MA_FAST).iloc[-1].ma
    return ma_slow, ma_fast


ma_slow, ma_fast = ma_caculate(klines)
print("慢速均线为%.2f,快速均线为%.2f" % (ma_slow, ma_fast))

while True:
    api.wait_update()
    # 每次k线更新，重新计算快慢均线
    if api.is_changing(klines.iloc[-1], "datetime"):
        ma_slow, ma_fast = ma_caculate(klines)
        print("慢速均线为%.2f,快速均线为%.2f" % (ma_slow, ma_fast))

    if api.is_changing(quote, "last_price"):
        # 开仓判断
        if position.pos_long == 0 and position.pos_short == 0:
            # 计算前后两根K线在当时K线范围波幅
            kl_range_cur = kline_range(-2)
            kl_range_pre = kline_range(-3)
            # 开多头判断，最近一根K线收盘价在短期均线和长期均线之上，前一根K线收盘价位于K线波动范围底部25%，最近这根K线收盘价位于K线波动范围顶部25%
            if klines.iloc[-2].close > max(ma_slow, ma_fast) and kl_range_pre <= 0.25 and kl_range_cur >= 0.75:
                print("最新价为:%.2f 开多头" % quote.last_price)
                target_pos.set_target_volume(100)

            # 开空头判断，最近一根K线收盘价在短期均线和长期均线之下，前一根K线收盘价位于K线波动范围顶部25%，最近这根K线收盘价位于K线波动范围底部25%
            elif klines.iloc[-2].close < min(ma_slow, ma_fast) and kl_range_pre >= 0.75 and kl_range_cur <= 0.25:
                print("最新价为:%.2f 开空头" % quote.last_price)
                target_pos.set_target_volume(-100)
            else:
                print("最新价位:%.2f ，未满足开仓条件" % quote.last_price)

        # 多头持仓止损策略
        elif position.pos_long > 0:
            # 在两根K线较低点减一跳，进行多头止损
            kline_low = min(klines.iloc[-2].low, klines.iloc[-3].low)
            if klines.iloc[-1].close <= kline_low - quote.price_tick:
                print("最新价为:%.2f,进行多头止损" % (quote.last_price))
                target_pos.set_target_volume(0)
            else:
                print("多头持仓，当前价格 %.2f,多头离场价格%.2f" %
                      (quote.last_price, kline_low - quote.price_tick))

        # 空头持仓止损策略
        elif position.pos_short > 0:
            # 在两根K线较高点加一跳，进行空头止损
            kline_high = max(klines.iloc[-2].high, klines.iloc[-3].high)
            if klines.iloc[-1].close >= kline_high + quote.price_tick:
                print("最新价为:%.2f 进行空头止损" % quote.last_price)
                target_pos.set_target_volume(0)
            else:
                print("空头持仓，当前价格 %.2f,空头离场价格%.2f" %
                      (quote.last_price, kline_high + quote.price_tick))



#!/usr/bin/env python
#  -*- coding: utf-8 -*-
__author__ = 'limin'

from tqsdk import TqApi, TqAuth
from tqsdk.ta import MA

'''
画图示例: 在附图中画指标线
(将画图代码放在循环中即可使图像随着行情推进而更新)
注意: 画图示例中用到的数据不含有实际意义，请根据自己的实际策略情况进行修改
'''

api = TqApi(web_gui=":9878", auth=TqAuth("快期账户", "账户密码"))  # web_gui="[ip]:port", 指定 web 界面地址的 ip 和 port
klines = api.get_kline_serial("SHFE.rb2105", 24 * 60 * 60)

while True:
    # 将画图代码放在循环中即可使图像随着行情推进而更新
    ma = MA(klines, 30)  # 使用tqsdk自带指标函数计算均线

    # 示例1: 在附图中画一根绿色的ma指标线
    klines["ma_B2"] = ma.ma
    klines["ma_B2.board"] = "B2"  # 设置附图: 可以设置任意字符串,同一字符串表示同一副图
    klines["ma_B2.color"] = "green"  # 设置为绿色. 以下设置颜色方式都可行: "green", "#00FF00", "rgb(0,255,0)", "rgba(0,125,0,0.5)"

    # 示例2: 在另一个附图画一根比ma小4的宽度为4的紫色指标线
    klines["ma_4"] = ma.ma - 4
    klines["ma_4.board"] = "MA4"  # 设置为另一个附图
    klines["ma_4.color"] = 0xFF9933CC  # 设置为紫色, 或者 "#9933FF"
    klines["ma_4.width"] = 4  # 设置宽度为4，默认为1

    api.wait_update()



#!/usr/bin/env python
#  -*- coding: utf-8 -*-
__author__ = 'chengzhi'

from datetime import datetime
from contextlib import closing
from tqsdk import TqApi, TqAuth
from tqsdk.tools import DataDownloader

api = TqApi(auth=TqAuth("快期账户", "账户密码"))
# 下载从 2018-01-01凌晨6点 到 2018-06-01下午4点 的 cu1805 分钟线数据
kd = DataDownloader(api, symbol_list="SHFE.cu1805", dur_sec=60,
                    start_dt=datetime(2018, 1, 1, 6, 0 ,0), end_dt=datetime(2018, 6, 1, 16, 0, 0), csv_file_name="kline.csv")
# 下载从 2018-05-01凌晨0点 到 2018-07-01凌晨0点 的 T1809 盘口Tick数据
td = DataDownloader(api, symbol_list="CFFEX.T1809", dur_sec=0,
                    start_dt=datetime(2018, 5, 1), end_dt=datetime(2018, 7, 1), csv_file_name="tick.csv")
# 使用with closing机制确保下载完成后释放对应的资源
with closing(api):
    while not kd.is_finished() or not td.is_finished():
        api.wait_update()
        print("progress: kline: %.2f%% tick:%.2f%%" % (kd.get_progress(), td.get_progress()))



#!/usr/bin/env python
#  -*- coding: utf-8 -*-
__author__ = 'chengzhi'

import datetime
from tqsdk import TqApi, TqAuth
from tqsdk.ta import *

api = TqApi(auth=TqAuth("快期账户", "账户密码"))
# 获得 cu1909 10秒K线的引用
klines = api.get_kline_serial("SHFE.cu1910", 10, data_length=3000)

print("K线时间", datetime.datetime.fromtimestamp(klines.iloc[-1]["datetime"] / 1e9))
print(klines)

print("ATR", ATR(klines, 26))
print("BIAS", BIAS(klines, 6))
print("BOLL", BOLL(klines, 3, 5))
print("DMI", DMI(klines, 14, 6))
print("KDJ", KDJ(klines, 9, 3, 3))
print("MA", MA(klines, 3))
print("MACD", MACD(klines, 20, 35, 10))
print("SAR", SAR(klines, 4, 0.02, 0.2))

api.close()



#!/usr/bin/env python
#  -*- coding: utf-8 -*-
__author__ = 'mayanqiong'

from shinny_structlog import ShinnyLoggerAdapter

from tqsdk.baseModule import TqModule
from tqsdk.channel import TqChan
from tqsdk.connect import TqConnect, TsReconnectHandler

class TqTradingStatus(TqModule):
    """
    交易状态模块，建立 websocket 连接
    """

    async def _run(self, api, api_send_chan, api_recv_chan, md_send_chan, md_recv_chan):
        self._api = api
        self._logger = self._api._logger.getChild("TradingStatus")
        self._init_ts_ws = False
        self._api_send_chan = api_send_chan
        self._api_recv_chan = api_recv_chan
        self._md_send_chan = md_send_chan
        self._md_recv_chan = md_recv_chan
        self._ts_send_chan = TqChan(self._api, chan_name="send to ts_reconn")
        self._ts_recv_chan = TqChan(self._api, chan_name="recv from ts_reconn")
        await super(TqTradingStatus, self)._run(api, api_send_chan, api_recv_chan, md_send_chan, md_recv_chan, self._ts_send_chan, self._ts_recv_chan)

    async def _handle_recv_data(self, pack, chan):
        """
        处理所有上游收到的数据包
        """
        if pack['aid'] == 'rtn_data':
            if chan == self._md_recv_chan:  # 从行情收到的数据包
                self._diffs.extend(pack.get('data', []))
            elif chan == self._ts_recv_chan:  # 从交易状态服务收到的数据包
                diffs = pack.get('data', [])
                for d in diffs:
                    for symbol, ts in d.get('trading_status', {}).items():
                        if ts['trade_status'] not in ["AUCTIONORDERING", "CONTINOUS"]:
                            ts['trade_status'] = "NOTRADING"
                self._diffs.extend(diffs)
        else:
            await self._api_recv_chan.send(pack)

    async def _handle_req_data(self, pack):
        """处理所有下游发送的非 peek_message 数据包"""
        if pack['aid'] == 'subscribe_trading_status':
            if self._init_ts_ws is False:
                self._init_ts_ws = True
                self._create_ts_run()
            await self._ts_send_chan.send(pack)
        else:
            await self._md_send_chan.send(pack)

    def _create_ts_run(self):
        ts_url = "wss://trading-status.shinnytech.com/status"
        conn_logger = self._api._logger.getChild("TqConnect")
        ws_ts_send_chan = TqChan(self._api, chan_name="send to ts")
        ws_ts_recv_chan = TqChan(self._api, chan_name="recv from ts")
        ws_ts_send_chan._logger_bind(chan_from="ts_reconn", url=ts_url)
        ws_ts_recv_chan._logger_bind(chan_to="ts_reconn", url=ts_url)
        conn = TqConnect(logger=ShinnyLoggerAdapter(conn_logger, url=ts_url), conn_id="ts")
        self._api.create_task(conn._run(self._api, ts_url, ws_ts_send_chan, ws_ts_recv_chan), _caller_api=True)
        ts_reconnect = TsReconnectHandler(logger=ShinnyLoggerAdapter(self._logger.getChild("TsReconnect"), url=ts_url))
        self._ts_send_chan._logger_bind(chan_from="ts", url=ts_url)
        self._ts_recv_chan._logger_bind(chan_to="ts", url=ts_url)
        self._api.create_task(ts_reconnect._run(self._api, self._ts_send_chan, self._ts_recv_chan, ws_ts_send_chan, ws_ts_recv_chan), _caller_api=True)



#!/usr/bin/env python
#  -*- coding: utf-8 -*-
__author__ = 'yanqiong'

from tqsdk import TqApi, TqAuth, TargetPosTask

'''
连续3根阴线就做空，连续3根阳线就做多，否则空仓
'''

api = TqApi(auth=TqAuth("快期账户", "账户密码"))
# 设定连续多少根阳线/阴线
length = 3
# 获得 ni2205 10秒K线的引用, 长度为 length+1
klines = api.get_kline_serial("SHFE.ni2205", 10, data_length=length + 1)
# 创建 ni2205 的目标持仓 task，该 task 负责调整 ni2105 的仓位到指定的目标仓位, offset_priority的用法详见文档
target_pos = TargetPosTask(api, "SHFE.ni2205", offset_priority="今昨开")

while True:
    api.wait_update()
    # 只有在新创建出K线时才判断开平仓条件
    if api.is_changing(klines.iloc[-1], "datetime"):
        # 跳过最后一根刚生成的K线
        df = klines.iloc[:-1]
        # 比较收盘价和开盘价，判断是阳线还是阴线
        # df.close 为收盘价序列, df.open 为开盘价序列, ">"(pandas.Series.gt) 返回收盘价是否大于开盘价的一个新序列
        up = df.close > df.open
        down = df.close < df.open
        if all(up):
            print("连续阳线: 目标持仓 多头1手")
            # 设置目标持仓为正数表示多头，负数表示空头，0表示空仓
            target_pos.set_target_volume(1)
        elif all(down):
            print("连续阴线: 目标持仓 空头1手")
            target_pos.set_target_volume(-1)
        else:
            print("目标持仓: 空仓")
            target_pos.set_target_volume(0)



from tqsdk import TqApi, TqAuth
from tqsdk.ta import VOLATILITY_CURVE

api = TqApi(auth=TqAuth("快期账户", "账户密码"))

# 获取 m2112 的看跌期权
underlying = "DCE.m2101"
options = api.query_options(underlying_symbol=underlying, option_class="PUT", expired=False)

# 批量获取合约的行情信息, 存储结构必须为 dict, key 为合约, value 为行情数据
quote = {}
for symbol in options:
    quote[symbol] = api.get_quote(symbol)
options.append(underlying)

klines = api.get_kline_serial(options, 24 * 60 * 60, 20)

# 使用 VOLATILITY_CURVE 函数计算波动率曲面
vc = VOLATILITY_CURVE(klines, quote, underlying, r=0.025)

print(vc)

api.close()



#!/usr/bin/env python
#  -*- coding: utf-8 -*-
__author__ = 'limin'

'''
Volume Weighted Average Price策略 (难度：高级)
参考: https://www.shinnytech.com/blog/vwap
注: 该示例策略仅用于功能示范, 实盘时请根据自己的策略/经验进行修改
'''

import datetime
from tqsdk import TqApi, TqAuth, TargetPosTask

TIME_CELL = 5 * 60  # 等时长下单的时间单元, 单位: 秒
TARGET_VOLUME = 300  # 目标交易手数 (>0: 多头, <0: 空头)
SYMBOL = "DCE.jd2001"  # 交易合约代码
HISTORY_DAY_LENGTH = 20  # 使用多少天的历史数据用来计算每个时间单元的下单手数
START_HOUR, START_MINUTE = 9, 35  # 计划交易时段起始时间点
END_HOUR, END_MINUTE = 10, 50  # 计划交易时段终点时间点

api = TqApi(auth=TqAuth("快期账户", "账户密码"))
print("策略开始运行")
# 根据 HISTORY_DAY_LENGTH 推算出需要订阅的历史数据长度, 需要注意history_day_length与time_cell的比例关系以避免超过订阅限制
time_slot_start = datetime.time(START_HOUR, START_MINUTE)  # 计划交易时段起始时间点
time_slot_end = datetime.time(END_HOUR, END_MINUTE)  # 计划交易时段终点时间点
klines = api.get_kline_serial(SYMBOL, TIME_CELL, data_length=int(10 * 60 * 60 / TIME_CELL * HISTORY_DAY_LENGTH))
target_pos = TargetPosTask(api, SYMBOL)
position = api.get_position(SYMBOL)  # 持仓信息


def get_kline_time(kline_datetime):
    """获取k线的时间(不包含日期)"""
    kline_time = datetime.datetime.fromtimestamp(kline_datetime // 1000000000).time()  # 每根k线的时间
    return kline_time


def get_market_day(kline_datetime):
    """获取k线所对应的交易日"""
    kline_dt = datetime.datetime.fromtimestamp(kline_datetime // 1000000000)  # 每根k线的日期和时间
    if kline_dt.hour >= 18:  # 当天18点以后: 移到下一个交易日
        kline_dt = kline_dt + datetime.timedelta(days=1)
    while kline_dt.weekday() >= 5:  # 是周六或周日,移到周一
        kline_dt = kline_dt + datetime.timedelta(days=1)
    return kline_dt.date()


# 添加辅助列: time及date, 分别为K线时间的时:分:秒和其所属的交易日
klines["time"] = klines.datetime.apply(lambda x: get_kline_time(x))
klines["date"] = klines.datetime.apply(lambda x: get_market_day(x))

# 获取在预设交易时间段内的所有K线, 即时间位于 time_slot_start 到 time_slot_end 之间的数据
if time_slot_end > time_slot_start:  # 判断是否类似 23:00:00 开始， 01:00:00 结束这样跨天的情况
    klines = klines[(klines["time"] >= time_slot_start) & (klines["time"] <= time_slot_end)]
else:
    klines = klines[(klines["time"] >= time_slot_start) | (klines["time"] <= time_slot_end)]

# 由于可能有节假日导致部分天并没有填满整个预设交易时间段
# 因此去除缺失部分交易时段的日期(即剩下的每个日期都包含预设的交易时间段内所需的全部时间单元)
date_cnt = klines["date"].value_counts()
max_num = date_cnt.max()  # 所有日期中最完整的交易时段长度
need_date = date_cnt[date_cnt == max_num].sort_index().index[-HISTORY_DAY_LENGTH - 1:-1]  # 获取今天以前的预设数目个交易日的日期
df = klines[klines["date"].isin(need_date)]  # 最终用来计算的k线数据

# 计算每个时间单元的成交量占比, 并使用算数平均计算出预测值
datetime_grouped = df.groupby(['date', 'time'])['volume'].sum()  # 将K线的volume按照date、time建立多重索引分组
# 计算每个交易日内的预设交易时间段内的成交量总和(level=0: 表示按第一级索引"data"来分组)后,将每根k线的成交量除以所在交易日内的总成交量,计算其所占比例
volume_percent = datetime_grouped / datetime_grouped.groupby(level=0).sum()
predicted_percent = volume_percent.groupby(level=1).mean()  # 将历史上相同时间单元的成交量占比使用算数平均计算出预测值
print("各时间单元成交量占比: %s" % predicted_percent)

# 计算每个时间单元的成交量预测值
predicted_volume = {}  # 记录每个时间单元需调整的持仓量
percentage_left = 1  # 剩余比例
volume_left = TARGET_VOLUME  # 剩余手数
for index, value in predicted_percent.items():
    volume = round(volume_left * (value / percentage_left))
    predicted_volume[index] = volume
    percentage_left -= value
    volume_left -= volume
print("各时间单元应下单手数: %s" % predicted_volume)

# 交易
current_volume = 0  # 记录已调整持仓量
while True:
    api.wait_update()
    # 新产生一根K线并且在计划交易时间段内: 调整目标持仓量
    if api.is_changing(klines.iloc[-1], "datetime"):
        t = datetime.datetime.fromtimestamp(klines.iloc[-1]["datetime"] // 1000000000).time()
        if t in predicted_volume:
            current_volume += predicted_volume[t]
            print("到达下一时间单元,调整持仓为: %d" % current_volume)
            target_pos.set_target_volume(current_volume)
    # 用持仓信息判断是否完成所有目标交易手数
    if api.is_changing(position, "volume_long") or api.is_changing(position, "volume_short"):
        if position["volume_long"] - position["volume_short"] == TARGET_VOLUME:
            break

api.close()



#!/usr/bin/env python
#  -*- coding: utf-8 -*-
__author__ = 'chengzhi'

from tqsdk import TqApi, TqAuth, TargetPosTask

'''
如果当前价格大于10秒K线的MA15则开多仓 (使用 TargetPosTask 调仓工具)
如果小于则平仓
'''
api = TqApi(auth=TqAuth("快期账户", "账户密码"))
# 获得 m2207 10秒K线的引用
klines = api.get_kline_serial("DCE.m2207", 10)
# 创建 m2207 的目标持仓 task，该 task 负责调整 m2207 的仓位到指定的目标仓位
target_pos = TargetPosTask(api, "DCE.m2207")

while True:
    api.wait_update()
    if api.is_changing(klines):
        ma = sum(klines.close.iloc[-15:]) / 15
        print("最新价", klines.close.iloc[-1], "MA", ma)
        if klines.close.iloc[-1] > ma:
            print("最新价大于MA: 目标多头5手")
            # 设置目标持仓为多头5手
            target_pos.set_target_volume(5)
        elif klines.close.iloc[-1] < ma:
            print("最新价小于MA: 目标空仓")
            # 设置目标持仓为空仓
            target_pos.set_target_volume(0)



#!/usr/bin/env python
#  -*- coding: utf-8 -*-
__author__ = 'mayanqiong'

import asyncio
from typing import Optional, Union

from pandas import DataFrame

from tqsdk.api import TqApi, TqAccount, TqSim, TqKq
from tqsdk.channel import TqChan
from tqsdk.datetime import _get_trade_timestamp
from tqsdk.lib.target_pos_task import TargetPosTask
from tqsdk.lib.utils import _check_time_table, _get_deadline_from_interval
from tqsdk.objs import Trade


class TargetPosScheduler(object):
    """算法执行引擎，根据设定的目标持仓任务列表，调整指定合约到目标头寸"""

    def __init__(self, api: TqApi, symbol: str, time_table: DataFrame, offset_priority: str = "今昨,开",
                 min_volume: Optional[int] = None, max_volume: Optional[int] = None, trade_chan: Optional[TqChan] = None,
                 trade_objs_chan: Optional[TqChan] = None, account: Optional[Union[TqAccount, TqKq, TqSim]] = None) -> None:
        """
        创建算法执行引擎实例，根据设定的目标持仓任务列表，调用 TargetPosTask 来调整指定合约到目标头寸。

        **注意:**
            1. TargetPosScheduler 创建后不会立即不下单或撤单, 它的下单和撤单动作, 是在之后的每次 wait_update 时执行的. 因此, **需保证后续还会调用wait_update()** 。

            2. 请勿同时使用 TargetPosScheduler、TargetPosTask、insert_order() 函数, 否则将导致报错或错误下单。

            3. `symbol`，`offset_priority`，`min_volume`，`max_volume`，`trade_chan`，`trade_objs_chan`，`account` 这几个参数会直接传给 TargetPosTask，请按照 TargetPosTask 的说明设置参数。

        Args:
            api (TqApi): TqApi实例，该task依托于指定api下单/撤单

            symbol (str): 负责调整的合约代码

            time_table (DataFrame): 目标持仓任务列表，每一行表示一项目标持仓任务，其应该包含以下几列：
                + interval: 当前这项任务的持续时间长度，单位为秒，经过这么多秒之后，此项任务应该退出，剩余未调整到的目标持仓，会留到下一项任务中
                    * 注意1：对于最后一项任务，会按照当前项参数，调整到目标持仓后立即退出（时间参数不对最后一项任务起作用）
                    * 注意2：时间长度可以跨非交易时间段（可以跨小节等待），但是不可以跨交易日
                + target_pos: 当前这项任务的目标净持仓手数
                + price: 当前这项任务的下单价格模式，此列中非 None 的项，会作为创建 TargetPosTask 实例的 price 参数，支持以下几种参数：
                    * None: 不下单，表示暂停一段时间
                    * "PASSIVE" : 排队价下单
                    * "ACTIVE": 对价下单
                    * Callable (direction: str) -> Union[float, int]: 传入函数作为价格参数，函数参数为下单方向，函数返回值是下单价格。如果返回 nan，程序会抛错。

            offset_priority (str): [可选]开平仓顺序，昨=平昨仓，今=平今仓，开=开仓，逗号=等待之前操作完成

                                   对于下单指令区分平今/昨的交易所(如上期所)，按照今/昨仓的数量计算是否能平今/昨仓
                                   对于下单指令不区分平今/昨的交易所(如中金所)，按照“先平当日新开仓，再平历史仓”的规则计算是否能平今/昨仓，如果这些交易所设置为"昨开"在有当日新开仓和历史仓仓的情况下，会自动跳过平昨仓进入到下一步

                                   * "今昨,开" 表示先平今仓，再平昨仓，等待平仓完成后开仓，对于没有单向大边的品种避免了开仓保证金不足
                                   * "今昨开" 表示先平今仓，再平昨仓，并开仓，所有指令同时发出，适合有单向大边的品种
                                   * "昨开" 表示先平昨仓，再开仓，禁止平今仓，适合股指这样平今手续费较高的品种
                                   * "开" 表示只开仓，不平仓，适合需要进行锁仓操作的品种

            min_volume (int): [可选] **大单拆分模式下** 每笔最小下单的手数，默认不启用 **大单拆分模式**

            max_volume (int): [可选] **大单拆分模式下** 每笔最大下单的手数，默认不启用 **大单拆分模式**

            trade_chan (TqChan): [可选]成交通知channel, 当有成交发生时会将成交手数(多头为正数，空头为负数)发到该channel上

            trade_objs_chan (TqChan): [可选]成交对象通知channel, 当有成交发生时会将成交对象发送到该channel上

            account (TqAccount/TqKq/TqSim): [可选]指定发送下单指令的账户实例, 多账户模式下，该参数必须指定

        Example::

            from pandas import DataFrame
            from tqsdk import TqApi, TargetPosScheduler

            api = TqApi(auth=TqAuth("快期账户", "账户密码"))
            time_table = DataFrame([
                [25, 10, "PASSIVE"],
                [5, 10, "ACTIVE"],
                [25, 20, "PASSIVE"],
                [5, 20, "ACTIVE"],
            ], columns=['interval', 'target_pos', 'price'])

            scheduler = TargetPosScheduler(api, 'SHFE.cu2112', time_table=time_table)
            while True:
                api.wait_update()
                if scheduler.is_finished():
                    break

            print("打印出 scheduler 全部成交以及成交均价")
            print(scheduler.trades_df)
            average_trade_price = sum(scheduler.trades_df['price'] * scheduler.trades_df['volume']) / sum(scheduler.trades_df['volume'])
            print("成交均价:", average_trade_price)
            api.close()
        """
        self._api = api
        self._account = api._account._check_valid(account)

        # 这些参数直接传给 TargetPosTask，由 TargetPosTask 来检查其合法性
        self._symbol = symbol
        self._offset_priority = offset_priority
        self._min_volume = min_volume
        self._max_volume = max_volume
        self._trade_chan = trade_chan

        self._trade_objs_chan = trade_objs_chan if trade_objs_chan else TqChan(self._api)
        self._time_table = _check_time_table(time_table)
        self._task = self._api.create_task(self._run())

        self._trade_keys = list(Trade(None).keys())
        self.trades_df = DataFrame(columns=self._trade_keys)  # 所有的 trade 列表
        self._trade_recv_task = self._api.create_task(self._trade_recv())

    async def _run(self):
        """负责调整目标持仓的task"""
        quote = await self._api.get_quote(self._symbol)
        self._time_table['deadline'] = _get_deadline_from_interval(quote, self._time_table['interval'])
        target_pos_task = None
        try:
            _index = 0  # _index 表示下标
            for index, row in self._time_table.iterrows():
                if row['price'] is None:
                    target_pos_task = None
                else:
                    target_pos_task = TargetPosTask(api=self._api, symbol=self._symbol, price=row['price'],
                                                    offset_priority=self._offset_priority,
                                                    min_volume=self._min_volume, max_volume=self._max_volume,
                                                    trade_chan=self._trade_chan,
                                                    trade_objs_chan=self._trade_objs_chan,
                                                    account=self._account)
                    target_pos_task.set_target_volume(row['target_pos'])
                if _index < self._time_table.shape[0] - 1:  # 非最后一项
                    async for _ in self._api.register_update_notify(quote):
                        if _get_trade_timestamp(quote.datetime, float('nan')) > row['deadline']:
                            if target_pos_task:
                                target_pos_task._task.cancel()
                                await asyncio.gather(target_pos_task._task, return_exceptions=True)
                            break
                elif target_pos_task:  # 最后一项，如果有 target_pos_task 等待持仓调整完成，否则直接退出
                    position = self._account.get_position(self._symbol)
                    async for _ in self._api.register_update_notify(position):
                        if position.pos == row['target_pos']:
                            break
                _index = _index + 1
        finally:
            if target_pos_task:
                target_pos_task._task.cancel()
                await asyncio.gather(target_pos_task._task, return_exceptions=True)
            await self._trade_objs_chan.close()
            await self._trade_recv_task

    async def _trade_recv(self):
        async for trade in self._trade_objs_chan:
            self.trades_df.loc[self.trades_df.shape[0]] = [trade[k] for k in self._trade_keys]

    def cancel(self):
        """
        取消当前 TargetPosScheduler 实例，会将该实例已经发出但还是未成交的委托单撤单。

        Example::

            from pandas import DataFrame
            from tqsdk import TqApi, TargetPosScheduler

            api = TqApi(auth=TqAuth("快期账户", "账户密码"))
            time_table = DataFrame([
                [25, 10, "PASSIVE"],
                [5, 10, "ACTIVE"],
                [25, 20, "PASSIVE"],
                [5, 20, "ACTIVE"],
            ], columns=['interval', 'target_pos', 'price'])

            scheduler = TargetPosScheduler(api, 'SHFE.cu2112', time_table=time_table)

            api.wait_update()
            # 运行代码。。。
            scheduler.cancel()
            while True:
                api.wait_update()
                if scheduler.is_finished():
                    break
            api.close()
        """
        if not self._task.done():
            self._task.cancel()
        return

    def is_finished(self):
        """
        返回当前 TargetPosScheduler 实例是否已经结束。即此实例不会再发出下单或者撤单的任何动作。

        Returns:
            bool: 当前 TargetPosScheduler 实例是否已经结束
        """
        return self._task.done()



#!usr/bin/env python3
# -*- coding:utf-8 -*-
__author__ = 'yanqiong'


from typing import Optional, Union

from tqsdk.diff import _get_obj
from tqsdk.entity import Entity
from tqsdk.objs import Account, Order, Trade, Position, SecurityAccount, SecurityOrder, SecurityTrade, SecurityPosition


def _get_api_instance(self):
    if hasattr(self, '_api'):
        return self._api
    import inspect
    raise Exception(f"未初始化 TqApi。请在 api 初始化后调用 {inspect.stack()[1].function}。")


class FutureMixin:

    _account_type = "FUTURE"

    def get_account(self) -> Account:
        """
        获取用户账户资金信息

        Returns:
            :py:class:`~tqsdk.objs.Account`: 返回一个账户对象引用. 其内容将在 :py:meth:`~tqsdk.api.TqApi.wait_update` 时更新

        Example1::

            # 获取当前浮动盈亏
            from tqsdk import TqApi, TqAuth

            tqacc = TqAccount("N南华期货", "123456", "123456")
            api = TqApi(account=tqacc, auth=TqAuth("快期账户", "账户密码"))
            account = tqacc.get_account()
            print(account.float_profit)

            # 预计的输出是这样的:
            2180.0
            ...

        Example2::

            # 多账户模式下, 分别获取各账户浮动盈亏
            from tqsdk import TqApi, TqAuth, TqMultiAccount, TqAccount, TqKq, TqSim

            account = TqAccount("N南华期货", "123456", "123456")
            tqkq = TqKq()
            tqsim = TqSim()
            api = TqApi(TqMultiAccount([account, tqkq, tqsim]), auth=TqAuth("快期账户", "账户密码"))
            account1 = account.get_account()
            account2 = tqkq.get_account()
            account3 = tqsim.get_account()
            print(f"账户 1 浮动盈亏 {account1.float_profit}, 账户 2 浮动盈亏 {account2.float_profit}, 账户 3 浮动盈亏 {account3.float_profit}")
            api.close()

        """
        api = _get_api_instance(self)
        return _get_obj(api._data, ["trade", self._account_key, "accounts", "CNY"], Account(api))

    def get_position(self, symbol: Optional[str] = None) -> Union[Position, Entity]:
        """
        获取用户持仓信息

        Args:
            symbol (str): [可选]合约代码, 不填则返回所有持仓

        Returns:
            :py:class:`~tqsdk.objs.Position`: 当指定了 symbol 时, 返回一个持仓对象引用。
            其内容将在 :py:meth:`~tqsdk.api.TqApi.wait_update` 时更新。

            不填 symbol 参数调用本函数, 将返回包含用户所有持仓的一个 ``tqsdk.objs.Entity`` 对象引用, 使用方法与dict一致, \
            其中每个元素的 key 为合约代码, value 为 :py:class:`~tqsdk.objs.Position`。

            注意: 为保留一些可供用户查询的历史信息, 如 volume_long_yd(本交易日开盘前的多头持仓手数) 等字段, 因此服务器会返回当天已平仓合约( pos_long 和 pos_short 等字段为0)的持仓信息

        Example1::

            # 获取 DCE.m2109 当前浮动盈亏
            from tqsdk import TqApi, TqAuth, TqAccount

            tqacc = TqAccount("N南华期货", "123456", "123456")
            api = TqApi(account=tqacc, auth=TqAuth("快期账户", "账户密码"))
            position = tqacc.get_position("DCE.m2109")
            print(position.float_profit_long + position.float_profit_short)
            while api.wait_update():
                print(position.float_profit_long + position.float_profit_short)

            # 预计的输出是这样的:
            300.0
            330.0
            ...

        Example2::

            # 多账户模式下, 分别获取各账户浮动盈亏
            from tqsdk import TqApi, TqAuth, TqMultiAccount, TqAccount, TqKq, TqSim

            account = TqAccount("N南华期货", "123456", "123456")
            tqkq = TqKq()
            tqsim = TqSim()
            api = TqApi(TqMultiAccount([account, tqkq, tqsim]), auth=TqAuth("快期账户", "账户密码"))
            position1 = account.get_position("DCE.m2101")
            position2 = tqkq.get_position("DCE.m2101")
            position3 = tqsim.get_position("DCE.m2101")
            print(f"账户 1 'DCE.m2101' 浮动盈亏 {position1.float_profit_long + position1.float_profit_short}, ",
                  f"账户 2 'DCE.m2101' 浮动盈亏 {position2.float_profit_long + position2.float_profit_short}, ",
                  f"账户 3 'DCE.m2101' 浮动盈亏 {position3.float_profit_long + position3.float_profit_short}")
            api.close()

        """
        api = _get_api_instance(self)
        if symbol:
            return _get_obj(api._data, ["trade", self._account_key, "positions", symbol], Position(api))
        return _get_obj(api._data, ["trade", self._account_key, "positions"])

    def get_order(self, order_id: Optional[str] = None) -> Union[Order, Entity]:
        """
        获取用户委托单信息

        Args:
            order_id (str): [可选]单号, 不填单号则返回所有委托单

        Returns:
            :py:class:`~tqsdk.objs.Order`: 当指定了 order_id 时, 返回一个委托单对象引用。 \
            其内容将在 :py:meth:`~tqsdk.api.TqApi.wait_update` 时更新。

            不填 order_id 参数调用本函数, 将返回包含用户所有委托单的一个 ``tqsdk.objs.Entity`` 对象引用, \
            使用方法与dict一致, 其中每个元素的key为委托单号, value为 :py:class:`~tqsdk.objs.Order`

            注意: 在刚下单后, tqsdk 还没有收到回单信息时, 此对象中各项内容为空

        Example1::

            # 获取当前总挂单手数
            from tqsdk import TqApi, TqAuth

            tqacc = TqAccount("N南华期货", "123456", "123456")
            api = TqApi(account=tqacc, auth=TqAuth("快期账户", "账户密码"))
            orders = tqacc.get_order()
            while True:
                api.wait_update()
                print(sum(order.volume_left for oid, order in orders.items() if order.status == "ALIVE"))

            # 预计的输出是这样的:
            3
            3
            0
            ...

        Example2::

            # 多账户模式下, 分别获取各账户挂单手数
            from tqsdk import TqApi, TqAuth, TqMultiAccount, TqAccount, TqKq, TqSim

            account = TqAccount("N南华期货", "123456", "123456")
            tqkq = TqKq()
            tqsim = TqSim()
            api = TqApi(TqMultiAccount([account, tqkq, tqsim]), auth=TqAuth("快期账户", "账户密码"))
            orders1 = account.get_order()
            orders2 = tqkq.get_order()
            orders3 = tqsim.get_order()
            print(f"账户 1 挂单手数 {sum(order.volume_left for order in orders1.values() if order.status == "ALIVE")}, ",
                  f"账户 2 挂单手数 {sum(order.volume_left for order in orders2.values() if order.status == "ALIVE")}, ",
                  f"账户 3 挂单手数 {sum(order.volume_left for order in orders3.values() if order.status == "ALIVE")}")

            order = account.get_order(order_id="订单号")
            print(order)
            api.close()

        """
        api = _get_api_instance(self)
        if order_id:
            return _get_obj(api._data, ["trade", self._account_key, "orders", order_id], Order(api))
        return _get_obj(api._data, ["trade", self._account_key, "orders"])

    def get_trade(self, trade_id: Optional[str] = None) -> Union[Trade, Entity]:
        """
        获取用户成交信息

        Args:
            trade_id (str): [可选]成交号, 不填成交号则返回所有委托单

        Returns:
            :py:class:`~tqsdk.objs.Trade`: 当指定了trade_id时, 返回一个成交对象引用. \
            其内容将在 :py:meth:`~tqsdk.api.TqApi.wait_update` 时更新.

            不填trade_id参数调用本函数, 将返回包含用户当前交易日所有成交记录的一个tqsdk.objs.Entity对象引用, 使用方法与dict一致, \
            其中每个元素的key为成交号, value为 :py:class:`~tqsdk.objs.Trade`

            推荐优先使用 :py:meth:`~tqsdk.objs.Order.trade_records` 获取某个委托单的相应成交记录, 仅当确有需要时才使用本函数.

        Example::

            # 多账户模式下, 分别获取各账户的成交记录
            from tqsdk import TqApi, TqAuth, TqMultiAccount

            account = TqAccount("N南华期货", "123456", "123456")
            tqkq = TqKq()
            tqsim = TqSim()
            api = TqApi(TqMultiAccount([account, tqkq, tqsim]), auth=TqAuth("快期账户", "账户密码"))
            trades1 = account.get_trade()
            trades2 = tqkq.get_trade()
            trades3 = tqsim.get_trade()
            print(trades1)
            print(trades2)
            print(trades3)
            api.close()
        """
        api = _get_api_instance(self)
        if trade_id:
            return _get_obj(api._data, ["trade", self._account_key, "trades", trade_id], Trade(api))
        return _get_obj(api._data, ["trade", self._account_key, "trades"])


class StockMixin:

    _account_type = "STOCK"

    def get_account(self) -> SecurityAccount:
        """
        获取用户账户资金信息

        Returns:
            :py:class:`~tqsdk.objs.SecurityAccount`: 返回一个账户对象引用. 其内容将在 :py:meth:`~tqsdk.api.TqApi.wait_update` 时更新

        Example1::

            # 获取当前浮动盈亏
            from tqsdk import TqApi, TqAuth

            tqacc = TqAccount("N南华期货", "123456", "123456")
            api = TqApi(account=tqacc, auth=TqAuth("快期账户", "账户密码"))
            account = tqacc.get_account()
            print(account.float_profit)

            # 预计的输出是这样的:
            2180.0
            ...

        Example2::

            # 多账户模式下, 分别获取各账户浮动盈亏
            from tqsdk import TqApi, TqAuth, TqMultiAccount, TqAccount, TqKq, TqSim

            account = TqAccount("N南华期货", "123456", "123456")
            tqkq = TqKq()
            tqsim = TqSim()
            api = TqApi(TqMultiAccount([account, tqkq, tqsim]), auth=TqAuth("快期账户", "账户密码"))
            account1 = account.get_account()
            account2 = tqkq.get_account()
            account3 = tqsim.get_account()
            print(f"账户 1 浮动盈亏 {account1.float_profit}, 账户 2 浮动盈亏 {account2.float_profit}, 账户 3 浮动盈亏 {account3.float_profit}")
            api.close()

        """
        api = _get_api_instance(self)
        return _get_obj(api._data, ["trade", self._account_key, "accounts", "CNY"], SecurityAccount(api))

    def get_position(self, symbol: Optional[str] = None) -> Union[SecurityPosition, Entity]:
        """
        获取用户持仓信息

        Args:
            symbol (str): [可选]合约代码, 不填则返回所有持仓

        Returns:
            :py:class:`~tqsdk.objs.SecurityPosition`: 当指定了 symbol 时, 返回一个持仓对象引用。
            其内容将在 :py:meth:`~tqsdk.api.TqApi.wait_update` 时更新。

            不填 symbol 参数调用本函数, 将返回包含用户所有持仓的一个 ``tqsdk.objs.Entity`` 对象引用, 使用方法与dict一致, \
            其中每个元素的 key 为合约代码, value 为 :py:class:`~tqsdk.objs.SecurityPosition`。


        Example::

            from tqsdk import TqApi, TqAuth, TqKqStock
            tqkqstock = TqKqStock()
            api = TqApi(account=tqkqstock, auth=TqAuth("快期账户", "账户密码"))
            position = tqkqstock.get_position('SSE.10003624')
            print(f"建仓日期 {position.create_date}, 持仓数量 {position.volume}")
            api.close()

        """
        api = _get_api_instance(self)
        if symbol:
            return _get_obj(api._data, ["trade", self._account_key, "positions", symbol], SecurityPosition(api))
        return _get_obj(api._data, ["trade", self._account_key, "positions"])

    def get_order(self, order_id: Optional[str] = None) -> Union[SecurityOrder, Entity]:
        """
        获取用户委托单信息

        Args:
            order_id (str): [可选]单号, 不填单号则返回所有委托单

        Returns:
            :py:class:`~tqsdk.objs.SecurityOrder`: 当指定了 order_id 时, 返回一个委托单对象引用。 \
            其内容将在 :py:meth:`~tqsdk.api.TqApi.wait_update` 时更新。

            不填 order_id 参数调用本函数, 将返回包含用户所有委托单的一个 ``tqsdk.objs.Entity`` 对象引用, \
            使用方法与 dict 一致, 其中每个元素的 key 为委托单号, value为 :py:class:`~tqsdk.objs.SecurityOrder`

            注意: 在刚下单后, tqsdk 还没有收到回单信息时, 此对象中各项内容为空

        Example::

            from tqsdk import TqApi, TqAuth, TqKqStock
            tqkqstock = TqKqStock()
            api = TqApi(account=tqkqstock, auth=TqAuth("快期账户", "账户密码"))
            order = tqkqstock.get_order('委托单Id')
            print(f"委托股数 {order.volume_orign}, 剩余股数 {order.volume_left}")
            api.close()
        """
        api = _get_api_instance(self)
        if order_id:
            return _get_obj(api._data, ["trade", self._account_key, "orders", order_id], SecurityOrder(api))
        return _get_obj(api._data, ["trade", self._account_key, "orders"])

    def get_trade(self, trade_id: Optional[str] = None) -> Union[SecurityTrade, Entity]:
        """
        获取用户成交信息

        Args:
            trade_id (str): [可选]成交号, 不填成交号则返回所有委托单

        Returns:
            :py:class:`~tqsdk.objs.SecurityTrade`: 当指定了trade_id时, 返回一个成交对象引用. \
            其内容将在 :py:meth:`~tqsdk.api.TqApi.wait_update` 时更新.

            不填trade_id参数调用本函数, 将返回包含用户当前交易日所有成交记录的一个 ``tqsdk.objs.Entity`` 对象引用, 使用方法与dict一致, \
            其中每个元素的key为成交号, value为 :py:class:`~tqsdk.objs.SecurityTrade`

            推荐优先使用 :py:meth:`~tqsdk.objs.SecurityOrder.trade_records` 获取某个委托单的相应成交记录, 仅当确有需要时才使用本函数.

        Example::

            from tqsdk import TqApi, TqAuth, TqKqStock
            tqkqstock = TqKqStock()
            api = TqApi(account=tqkqstock, auth=TqAuth("快期账户", "账户密码"))
            trades = tqkqstock.get_trade('委托单Id')
            [print(trade.trade_id, f"成交股数 {trade.volume}, 成交价格 {trade.price}") for trade in trades]
            api.close()
        """
        api = _get_api_instance(self)
        if trade_id:
            return _get_obj(api._data, ["trade", self._account_key, "trades", trade_id], SecurityTrade(api))
        return _get_obj(api._data, ["trade", self._account_key, "trades"])



#:!/usr/bin/env python
#:  -*- coding: utf-8 -*-
__author__ = 'mayanqiong'

from collections import namedtuple
from datetime import datetime
from typing import Callable, Tuple

import aiohttp
from pandas import DataFrame, Series
from sgqlc.operation import Operation
from tqsdk.backtest import TqBacktest

from tqsdk.datetime import _get_expire_rest_days, _str_to_timestamp_nano
from tqsdk.ins_schema import ins_schema, _add_all_frags
from tqsdk.objs import Quote
from tqsdk.diff import _get_obj
from tqsdk.utils import _query_for_quote, _generate_uuid
from tqsdk.tafunc import _get_t_series, get_impv, _get_d1, get_delta, get_theta, get_gamma, get_vega, get_rho

"""
这两个类只在 api 中用到，主要为了支持用户异步中 await 
没有继承 Entity 类
"""


async def ensure_quote(api, quote):
    if quote.price_tick > 0 and quote.datetime != "":
        return quote
    async with api.register_update_notify(quote) as update_chan:
        async for _ in update_chan:
            if quote.price_tick > 0 and quote.datetime != "":
                return quote


async def ensure_quote_with_underlying(api, quote):
    await ensure_quote(api, quote)
    if quote.underlying_symbol:
        await ensure_quote(api, quote.underlying_quote)
    return quote


class QuoteList(list):
    """
    请求合约信息和行情信息，self._task 完成时，所有的合约已经收到了合约信息和行情信息
    """

    def __init__(self, api, quotes):
        self._api = api
        list.__init__(self, quotes)
        self._task = api.create_task(self._ensure_quotes(), _caller_api=True)
        for quote in quotes:
            # 为每个 quote 对象创建 _task
            if not hasattr(quote, '_task'):
                quote._task = api.create_task(ensure_quote_with_underlying(api, quote), _caller_api=True)

    async def _ensure_symbols(self):
        if all([q.price_tick > 0 for q in self]):
            return
        query_symbols = [q._path[-1] for q in self if not q.price_tick > 0]
        query_pack = _query_for_quote(query_symbols)
        self._api._send_pack(query_pack)
        async with self._api.register_update_notify(self) as update_chan:
            async for _ in update_chan:
                if all([q.price_tick > 0 for q in self]):
                    return

    async def _ensure_quotes(self):
        await self._ensure_symbols()
        self._api._auth._has_md_grants([q._path[-1] for q in self])  # 权限检查
        # 发送的请求会请求到所有字段，如果是期权也会请求标的的合约信息
        underlying_symbols = set([q.underlying_symbol for q in self if q.underlying_symbol])
        need_quotes = set([q._path[-1] for q in self]).union(underlying_symbols)
        if need_quotes - self._api._requests["quotes"] != set():
            self._api._requests["quotes"] = self._api._requests["quotes"].union(need_quotes)
            self._api._send_pack({
                "aid": "subscribe_quote",
                "ins_list": ",".join(self._api._requests["quotes"]),
            })
        if all([q.datetime != "" for q in self]):
            return self
        all_quotes = self + [_get_obj(self._api._data, ["quotes", s], self._api._prototype["quotes"]["#"]) for s in underlying_symbols]
        async with self._api.register_update_notify(self) as update_chan:
            async for _ in update_chan:
                if all([q.datetime != "" for q in all_quotes]):
                    return self

    def __await__(self):
        return self._task.__await__()


async def _query_graphql_async(api, query_id, query):
    api._send_pack({
        "aid": "ins_query",
        "query_id": query_id,
        "query": query
    })
    symbols = _get_obj(api._data, ["symbols"])
    async with api.register_update_notify(symbols) as update_chan:
        async for _ in update_chan:
            s = symbols.get(query_id, {})
            if s.get("query") == query:
                break


class SymbolList(list):
    """
    query 系列函数返回对象
    """

    def __init__(self, api, query_id: str, query: str, filter: Callable[[dict], list]):
        self._api = api
        self._query_id = query_id
        self._query = query
        self._filter = filter
        list.__init__(self, [])
        self._task = api.create_task(self._query_graphql(), _caller_api=True)

    async def _query_graphql(self):
        pack = {"query": self._query}
        symbols = _get_obj(self._api._data, ["symbols"])
        query_result = None
        for symbol in symbols.values():
            if symbol.items() >= pack.items():  # 检查是否发送过相同的请求
                query_result = symbol
        if query_result is None:
            await _query_graphql_async(self._api, self._query_id, self._query)
            query_result = symbols.get(self._query_id)
        self += self._filter(query_result)
        if isinstance(self._api._backtest, TqBacktest):  # 回测时，清空缓存的请求
            self._api._send_pack({
                "aid": "ins_query",
                "query_id": self._query_id,
                "query": ""
            })
        return self

    def __await__(self):
        return self._task.__await__()


class SymbolLevelList(namedtuple('SymbolLevel', ['in_money_options', 'at_money_options', 'out_of_money_options'])):
    """
    query 系列函数返回对象
    """

    def __new__(cls, *args, **kwargs):
        return super(SymbolLevelList, cls).__new__(cls, in_money_options=[], at_money_options=[], out_of_money_options=[])

    def __init__(self, api, query_id: str, query: str, filter: Callable[[dict], Tuple[list, list, list]]):
        self._api = api
        self._query_id = query_id
        self._query = query
        self._filter = filter
        self._task = api.create_task(self._query_graphql(), _caller_api=True)

    async def _query_graphql(self):
        pack = {"query": self._query}
        symbols = _get_obj(self._api._data, ["symbols"])
        query_result = None
        for symbol in symbols.values():
            if symbol.items() >= pack.items():  # 检查是否发送过相同的请求
                query_result = symbol
        if query_result is None:
            await _query_graphql_async(self._api, self._query_id, self._query)
            query_result = symbols.get(self._query_id)
        l0, l1, l2 = self._filter(query_result)
        self[0].extend(l0)
        self[1].extend(l1)
        self[2].extend(l2)
        if isinstance(self._api._backtest, TqBacktest):  # 回测时，清空缓存的请求
            self._api._send_pack({
                "aid": "ins_query",
                "query_id": self._query_id,
                "query": ""
            })
        return self

    def __await__(self):
        return self._task.__await__()


class TqDataFrame(DataFrame):

    def __init__(self, api, *args, **kwargs):
        super(TqDataFrame, self).__init__(*args, **kwargs)
        self.__dict__["_api"] = api
        self.__dict__["_task"] = api.create_task(self.async_update(), _caller_api=True)

    async def async_update(self):
        async with self._api.register_update_notify(self) as update_chan:
            async for _ in update_chan:
                if self._api._serials.get(id(self))["init"]:
                    return self

    def __await__(self):
        return self.__dict__["_task"].__await__()


class TqSymbolDataFrame(DataFrame):

    def __init__(self, api, symbol_list, backtest_timestamp, *args, **kwargs):
        self.__dict__["_api"] = api
        self.__dict__["_symbol_list"] = symbol_list
        self.__dict__["_backtest_timestamp"] = backtest_timestamp
        self.__dict__["_columns"] = [
            "ins_class",
            "instrument_id",
            "instrument_name",
            "price_tick",
            "volume_multiple",
            "max_limit_order_volume",
            "max_market_order_volume",
            "underlying_symbol",
            "strike_price",
            "exchange_id",
            "product_id",
            "expired",
            "expire_datetime",
            "expire_rest_days",
            "delivery_year",
            "delivery_month",
            "last_exercise_datetime",
            "exercise_year",
            "exercise_month",
            "option_class",
            "upper_limit",
            "lower_limit",
            "pre_settlement",
            "pre_open_interest",
            "pre_close",
            "trading_time_day",
            "trading_time_night"
        ]
        default_quote = Quote(None)
        data = [{k: (s if k == "instrument_id" else default_quote.get(k, None)) for k in self.__dict__["_columns"]} for s in symbol_list]
        super(TqSymbolDataFrame, self).__init__(data=data, columns=self.__dict__["_columns"], *args, **kwargs)
        self.__dict__["_task"] = api.create_task(self.async_update(), _caller_api=True)

    async def async_update(self):
        query_id = _generate_uuid("PYSDK_api")
        op = Operation(ins_schema.rootQuery)
        variables = {"instrument_id": self.__dict__["_symbol_list"]}
        if self.__dict__["_backtest_timestamp"]:
            variables["timestamp"] = self.__dict__["_backtest_timestamp"]
        query = op.multi_symbol_info(**variables)
        _add_all_frags(query)
        self.__dict__["_api"]._send_pack({
            "aid": "ins_query",
            "query_id": query_id,
            "query": op.__to_graphql__()
        })
        symbols = _get_obj(self.__dict__["_api"]._data, ["symbols"])
        async with self.__dict__["_api"].register_update_notify(symbols) as update_chan:
            async for _ in update_chan:
                query_result = symbols.get(query_id, {})
                if query_result:
                    all_keys = set(self.__dict__["_columns"])
                    all_keys.add('trading_time')
                    quotes = self.__dict__["_api"]._symbols_to_quotes(query_result, keys=all_keys)
                    self._quotes_to_dataframe(quotes)
                    if self.__dict__["_backtest_timestamp"]:
                        # 回测时这些字段应该为 nan
                        self.loc[:, ["upper_limit", "lower_limit", "pre_settlement", "pre_open_interest", "pre_close"]] = float('nan')
                        # 回测时清空请求，不缓存请求内容
                        self.__dict__["_api"]._send_pack({
                            "aid": "ins_query",
                            "query_id": query_id,
                            "query": ""
                        })
                    return self

    def _get_trading_time(self, quotes, symbol, key):
        v = quotes[symbol].get('trading_time', {'day': [], 'night': []}).get(key, [])
        return v if v else None

    def _quotes_to_dataframe(self, quotes):
        default_quote = Quote(None)
        for col in self.__dict__["_columns"]:
            if col == "expire_rest_days":
                current_dt = self._api._get_current_datetime().timestamp()
                self[col] = [_get_expire_rest_days(quotes[s]['expire_datetime'], current_dt)
                                    if quotes[s].get('expire_datetime') else float('nan')
                                    for s in self.__dict__["_symbol_list"]]
            elif col == "trading_time_day" or col == "trading_time_night":
                k = 'day' if col == "trading_time_day" else 'night'
                self[col] = Series([self._get_trading_time(quotes, s, k) for s in self.__dict__["_symbol_list"]])
            else:
                self[col] = Series([quotes[s].get(col, default_quote[col]) for s in self.__dict__["_symbol_list"]])

    def __await__(self):
        return self.__dict__["_task"].__await__()


class TqSymbolRankingDataFrame(DataFrame):

    def __init__(self, api, symbol, ranking_type, days, start_dt, broker):
        self.__dict__["_api"] = api
        params = {'symbol': symbol}
        if days is not None:
            params['days'] = days
        if start_dt is not None:
            params['start_date'] = start_dt.strftime("%Y%m%d")
        if broker is not None:
            params['broker'] = broker
        self.__dict__["_params"] = params
        self.__dict__["_symbol"] = symbol
        self.__dict__["_ranking_type"] = f"{ranking_type.lower()}_ranking"
        self.__dict__["_columns"] = [
            "datetime",
            "symbol",
            "exchange_id",
            "instrument_id",
            "broker",
            "volume",
            "volume_change",
            "volume_ranking",
            "long_oi",
            "long_change",
            "long_ranking",
            "short_oi",
            "short_change",
            "short_ranking"
        ]
        super(TqSymbolRankingDataFrame, self).__init__(data=[], columns=self.__dict__["_columns"])
        self.__dict__["_task"] = api.create_task(self.async_update(), _caller_api=True)

    async def _get_ranking_data(self, ranking_id):
        # 下载持仓排名数据，并将数据发回到 api.recv_chan
        async with aiohttp.ClientSession(headers=self.__dict__["_api"]._base_headers) as session:
            url = "https://symbol-ranking-system-fc-api.shinnytech.com/srs"
            async with session.get(url, params=self.__dict__["_params"]) as response:
                response.raise_for_status()
                content = await response.json()
                await self.__dict__["_api"]._ws_md_recv_chan.send({
                    "aid": "rtn_data",
                    "data": [{
                        "_symbol_rankings": {
                            ranking_id: content
                        },
                        "_symbol_rankings_finished": {
                            ranking_id: True
                        }
                    }]
                })

    async def async_update(self):
        await self.__dict__["_api"]._ensure_symbol_async(self.__dict__["_symbol"])
        ranking_id = _generate_uuid("PYSDK_rank")
        self.__dict__["_api"].create_task(self._get_ranking_data(ranking_id), _caller_api=True)  # 错误会抛给 api 处理
        symbol_rankings_finished = _get_obj(self.__dict__["_api"]._data, ["_symbol_rankings_finished"])
        async with self.__dict__["_api"].register_update_notify(symbol_rankings_finished) as update_chan:
            async for _ in update_chan:
                if not symbol_rankings_finished.get(ranking_id, False):
                    continue
                content = self.__dict__["_api"]._data.get("_symbol_rankings", {}).get(ranking_id, {})
                data = self._content_to_list(content)
                for i, d in enumerate(data):
                    self.loc[i] = d
                self.dropna(subset=[self.__dict__["_ranking_type"]], inplace=True)
                self.sort_values(by=['datetime', self.__dict__["_ranking_type"]], inplace=True, ignore_index=True)
                # 读完数据，清空数据
                await self.__dict__["_api"]._ws_md_recv_chan.send({
                    "aid": "rtn_data",
                    "data": [{
                        "_symbol_rankings": {
                            ranking_id: None
                        },
                        "_symbol_rankings_finished": {
                            ranking_id: None
                        }
                    }]
                })
                return self

    def _content_to_list(self, content):
        data = {}
        for dt in content.keys():
            for symbol in content[dt].keys():
                if content[dt][symbol] is None:
                    continue
                for data_type, rankings in content[dt][symbol].items():
                    for broker, rank_item in rankings.items():
                        item = data.setdefault((dt, symbol, broker), self._get_default_item(dt, symbol, broker))
                        if data_type == 'volume_ranking':
                            item['volume'] = rank_item['volume']
                            item['volume_change'] = rank_item['varvolume']
                            item['volume_ranking'] = rank_item['ranking']
                        elif data_type == 'long_ranking':
                            item['long_oi'] = rank_item['volume']
                            item['long_change'] = rank_item['varvolume']
                            item['long_ranking'] = rank_item['ranking']
                        elif data_type == 'short_ranking':
                            item['short_oi'] = rank_item['volume']
                            item['short_change'] = rank_item['varvolume']
                            item['short_ranking'] = rank_item['ranking']
        return data.values()

    def _get_default_item(self, dt, symbol, broker):
        return {
            "datetime": dt,
            "symbol": symbol,
            "exchange_id": symbol.split(".", maxsplit=1)[0],
            "instrument_id": symbol.split(".", maxsplit=1)[1],
            "broker": broker,
            "volume": float('nan'),
            "volume_change": float('nan'),
            "volume_ranking": float('nan'),
            "long_oi": float('nan'),
            "long_change": float('nan'),
            "long_ranking": float('nan'),
            "short_oi": float('nan'),
            "short_change": float('nan'),
            "short_ranking": float('nan')
        }

    def __await__(self):
        return self.__dict__["_task"].__await__()



class TqOptionGreeksDataFrame(DataFrame):

    def __init__(self, api, symbol_list, v_list, r):
        self.__dict__["_api"] = api
        self.__dict__["_symbol_list"] = symbol_list
        self.__dict__["_v_list"] = v_list
        self.__dict__["_r"] = r
        self.__dict__["_columns"] = [
            "instrument_id",
            "instrument_name",
            "option_class",
            "expire_rest_days",
            "expire_datetime",
            "underlying_symbol",
            "strike_price",
            "delta",
            "gamma",
            "theta",
            "vega",
            "rho"
        ]
        super(TqOptionGreeksDataFrame, self).__init__(data=[], columns=self.__dict__["_columns"])
        self.__dict__["_task"] = api.create_task(self.async_update(), _caller_api=True)

    async def async_update(self):
        symbol_list = self.__dict__["_symbol_list"]
        quotes = await self.__dict__["_api"].get_quote_list(symbol_list)
        if not all([q.ins_class.endswith("OPTION") for q in quotes]):
            raise Exception("quote 参数列表中元素必须是期权类型")
        for i, q in enumerate(quotes):
            self.loc[i] = {k: q.get(k, float('nan')) for k in self.__dict__["_columns"]}
        self._get_greeks(quotes)
        return self

    def _get_greeks(self, quotes):
        series_close = Series(data=[q.last_price for q in quotes])  # 期权最新价
        series_close1 = Series(data=[q.underlying_quote.last_price for q in quotes])  # 标的最新价
        series_o = Series(data=[q.option_class for q in quotes])
        series_datetime = Series(data=[_str_to_timestamp_nano(q.datetime) for q in quotes])
        series_expire_datetime = Series(data=[q.expire_datetime for q in quotes])
        series_t = _get_t_series(series_datetime, 0, series_expire_datetime)  # 到期时间
        if self.__dict__["_v_list"] is None:
            series_v = get_impv(series_close1, series_close, self["strike_price"], self.__dict__["_r"], 0.3, series_t, series_o)
        else:
            series_v = Series(data=self.__dict__["_v_list"])
        series_d1 = _get_d1(series_close1, self["strike_price"], self.__dict__["_r"], series_v, series_t)
        self["delta"] = get_delta(series_close1, self["strike_price"], self.__dict__["_r"], series_v, series_t, series_o, series_d1)
        self["theta"] = get_theta(series_close1, self["strike_price"], self.__dict__["_r"], series_v, series_t, series_o, series_d1)
        self["gamma"] = get_gamma(series_close1, self["strike_price"], self.__dict__["_r"], series_v, series_t, series_d1)
        self["vega"] = get_vega(series_close1, self["strike_price"], self.__dict__["_r"], series_v, series_t, series_d1)
        self["rho"] = get_rho(series_close1, self["strike_price"], self.__dict__["_r"], series_v, series_t, series_o, series_d1)

    def __await__(self):
        return self.__dict__["_task"].__await__()



#!/usr/bin/env python
#  -*- coding: utf-8 -*-
__author__ = 'yanqiong'

from tqsdk import TqApi, TqAuth
from tqsdk.ta import MACD

'''
画图示例: 在附图中画 macd 指标示例
注意: 画图示例中用到的数据不含有实际意义，请根据自己的实际策略情况进行修改
'''

def calc_macd_klines(klines):
    # 计算 macd 指标
    macd = MACD(klines, 12, 26, 9)
    # 用 K 线图模拟 MACD 指标柱状图
    klines["MACD.open"] = 0.0
    klines["MACD.close"] = macd["bar"]
    klines["MACD.high"] = klines["MACD.close"].where(klines["MACD.close"] > 0, 0)
    klines["MACD.low"] = klines["MACD.close"].where(klines["MACD.close"] < 0, 0)
    klines["MACD.board"] = "MACD"
    # 在 board=MACD 上添加 diff、dea 线
    klines["diff"] = macd["diff"]
    klines["diff.board"] = "MACD"
    klines["diff.color"] = "gray"
    klines["dea"] = macd["dea"]
    klines["dea.board"] = "MACD"
    klines["dea.color"] = "rgb(255,128,0)"


api = TqApi(auth=TqAuth("快期账户", "账户密码"), web_gui=True)
klines = api.get_kline_serial("SHFE.rb2105", 5*60, 200)
while True:
    calc_macd_klines(klines)
    api.wait_update()



#:!/usr/bin/env python
#:  -*- coding: utf-8 -*-
__author__ = 'chengzhi'

import copy
import json
import warnings

from tqsdk.diff import _get_obj
from tqsdk.entity import Entity
from tqsdk.utils import _query_for_init, _generate_uuid


class QuotesEntity(Entity):

    def __init__(self, api):
        self._api = api
        self._not_send_init_query = True

    def __iter__(self):
        message = """
            不推荐使用 api._data['quotes'] 获取全部合约，该使用方法会在 20201101 之后的版本中放弃维护。
            需要注意：
            * 在同步代码中，初次使用 api._data['quotes'] 获取全部合约会产生一个耗时很长的查询。
            * 在协程中，api._data['quotes'] 这种用法不支持使用。
            请尽快修改使用新的接口，参考链接 http://doc.shinnytech.com/tqsdk/reference/tqsdk.api.html#tqsdk.api.TqApi.query_quotes
        """
        warnings.warn(message, DeprecationWarning, stacklevel=3)
        self._api._logger.warning("deprecation", content="Deprecation Warning in api._data['quotes']")

        # 兼容旧版 tqsdk 所做的修改，来支持用户使用 for k,v in api._data.quotes.items() 类似的用法
        # 从 api._init_() 最后 3 行移到这里
        if self._not_send_init_query and self._api._stock:
            self._not_send_init_query = False
            q = _query_for_init()
            self._api.query_graphql(q, {}, _generate_uuid("PYSDK_quote"))
        return super().__iter__()


class Quote(Entity):
    """ Quote 是一个行情对象 """

    def __init__(self, api):
        self._api = api
        #: 行情从交易所发出的时间(北京时间), 格式为 "2017-07-26 23:04:21.000001"
        self.datetime: str = ""
        #: 卖一价
        self.ask_price1: float = float("nan")
        #: 卖一量
        self.ask_volume1: int = 0
        #: 买一价
        self.bid_price1: float = float("nan")
        #: 买一量
        self.bid_volume1: int = 0
        #: 卖二价
        self.ask_price2: float = float("nan")
        #: 卖二量
        self.ask_volume2: int = 0
        #: 买二价
        self.bid_price2: float = float("nan")
        #: 买二量
        self.bid_volume2: int = 0
        #: 卖三价
        self.ask_price3: float = float("nan")
        #: 卖三量
        self.ask_volume3: int = 0
        #: 买三价
        self.bid_price3: float = float("nan")
        #: 买三量
        self.bid_volume3: int = 0
        #: 卖四价
        self.ask_price4: float = float("nan")
        #: 卖四量
        self.ask_volume4: int = 0
        #: 买四价
        self.bid_price4: float = float("nan")
        #: 买四量
        self.bid_volume4: int = 0
        #: 卖五价
        self.ask_price5: float = float("nan")
        #: 卖五量
        self.ask_volume5: int = 0
        #: 买五价
        self.bid_price5: float = float("nan")
        #: 买五量
        self.bid_volume5: int = 0
        #: 最新价
        self.last_price: float = float("nan")
        #: 当日最高价
        self.highest: float = float("nan")
        #: 当日最低价
        self.lowest: float = float("nan")
        #: 开盘价
        self.open: float = float("nan")
        #: 收盘价
        self.close: float = float("nan")
        #: 当日均价
        self.average: float = float("nan")
        #: 成交量
        self.volume: int = 0
        #: 成交额
        self.amount: float = float("nan")
        #: 持仓量
        self.open_interest: int = 0
        #: 结算价
        self.settlement: float = float("nan")
        #: 涨停价
        self.upper_limit: float = float("nan")
        #: 跌停价
        self.lower_limit: float = float("nan")
        #: 昨持仓量
        self.pre_open_interest: int = 0
        #: 昨结算价
        self.pre_settlement: float = float("nan")
        #: 昨收盘价
        self.pre_close: float = float("nan")
        #: 合约价格变动单位
        self.price_tick: float = float("nan")
        #: 合约价格小数位数
        self.price_decs: int = 0
        #: 合约乘数
        self.volume_multiple: int = 0
        #: 最大限价单手数
        self.max_limit_order_volume: int = 0
        #: 最大市价单手数
        self.max_market_order_volume: int = 0
        #: 最小限价单手数
        self.min_limit_order_volume: int = 0
        #: 最小市价单手数
        self.min_market_order_volume: int = 0
        #: 标的合约
        self.underlying_symbol: str = ""
        #: 行权价
        self.strike_price: float = float("nan")
        #: 合约类型
        self.ins_class: str = ""
        #: 交易所内的合约代码
        self.instrument_id: str = ""
        #: 合约中文名
        self.instrument_name: str = ""
        #: 交易所代码
        self.exchange_id: str = ""
        #: 合约是否已下市
        self.expired: bool = False
        #: 交易时间段
        self.trading_time: TradingTime = TradingTime(self._api)
        #: 到期具体日，以秒为单位的 timestamp 值
        self.expire_datetime: float = float("nan")
        #: 期货交割日年份，只对期货品种有效。期权推荐使用最后行权日年份
        self.delivery_year: int = 0
        #: 期货交割日月份，只对期货品种有效。期权推荐使用最后行权日月份
        self.delivery_month: int = 0
        #: 期权最后行权日，以秒为单位的 timestamp 值
        self.last_exercise_datetime: float = float("nan")
        #: 期权最后行权日年份，只对期权品种有效。
        self.exercise_year: int = 0
        #: 期权最后行权日月份，只对期权品种有效。
        self.exercise_month: int = 0
        #: 期权方向
        self.option_class: str = ""
        #: 期权行权方式，美式:'A'，欧式:'E'
        self.exercise_type: str = ""
        #: 品种代码
        self.product_id: str = ""
        #: ETF实时单位基金净值
        self.iopv: float = float("nan")
        #: 日流通股数，只对证券产品有效。
        self.public_float_share_quantity: int = 0
        #: 除权表 ["20190601,0.15","20200107,0.2"…]
        self.stock_dividend_ratio: list = []
        #: 除息表 ["20190601,0.15","20200107,0.2"…]
        self.cash_dividend_ratio: list = []
        #: 距离到期日的剩余天数（自然日天数），正数表示距离到期日的剩余天数，0表示到期日当天，负数表示距离到期日已经过去的天数
        self.expire_rest_days: int = float('nan')

    def _instance_entity(self, path):
        super(Quote, self)._instance_entity(path)
        self.trading_time = copy.copy(self.trading_time)
        self.trading_time._instance_entity(path + ["trading_time"])

    @property
    def underlying_quote(self):
        """
        标的合约 underlying_symbol 所指定的合约对象，若没有标的合约则为 None

        :return: 标的指定的 :py:class:`~tqsdk.objs.Quote` 对象
        """
        if self.underlying_symbol:
            return self._api.get_quote(self.underlying_symbol)
        return None

    def __await__(self):
        assert self._task
        return self._task.__await__()


class TradingTime(Entity):
    """ TradingTime 是一个交易时间对象
        它不是一个可单独使用的类，而是用于定义 Quote 的 trading_time 字段的类型

        (每个连续的交易时间段是一个列表，包含两个字符串元素，分别为这个时间段的起止点)"""

    def __init__(self, api):
        self._api = api
        #: 白盘
        self.day: list = []
        #: 夜盘（注意：本字段中过了 24：00 的时间则在其基础往上加，如凌晨1点为 '25:00:00' ）
        self.night: list = []

    def __repr__(self):
        return json.dumps({"day": self.day, "night": self.night})


class TradingStatus(Entity):
    """ TradingStatus 是一个交易状态对象 """

    def __init__(self, api):
        self._api = api
        #: 合约
        self.symbol: str = ""
        #: 交易状态, AUCTIONORDERING: 集合竞价报单; CONTINOUS: 连续交易; NOTRADING: 非交易时段
        self.trade_status: str = ""

    def __await__(self):
        assert self._task
        return self._task.__await__()


class Kline(Entity):
    """ Kline 是一个K线对象 """

    def __init__(self, api):
        self._api = api
        #: K线起点时间(按北京时间)，自unix epoch(1970-01-01 00:00:00 GMT)以来的纳秒数
        self.datetime: int = 0
        #: K线起始时刻的最新价
        self.open: float = float("nan")
        #: K线时间范围内的最高价
        self.high: float = float("nan")
        #: K线时间范围内的最低价
        self.low: float = float("nan")
        #: K线结束时刻的最新价
        self.close: float = float("nan")
        #: K线时间范围内的成交量
        self.volume: int = 0
        #: K线起始时刻的持仓量
        self.open_oi: int = 0
        #: K线结束时刻的持仓量
        self.close_oi: int = 0


class Tick(Entity):
    """ Tick 是一个tick对象 """

    def __init__(self, api):
        self._api = api
        #: tick从交易所发出的时间(按北京时间)，自unix epoch(1970-01-01 00:00:00 GMT)以来的纳秒数
        self.datetime: int = 0
        #: 最新价
        self.last_price: float = float("nan")
        #: 当日均价
        self.average: float = float("nan")
        #: 当日最高价
        self.highest: float = float("nan")
        #: 当日最低价
        self.lowest: float = float("nan")
        #: 卖1价
        self.ask_price1: float = float("nan")
        #: 卖1量
        self.ask_volume1: int = 0
        #: 买1价
        self.bid_price1: float = float("nan")
        #: 买1量
        self.bid_volume1: int = 0
        #: 卖2价
        self.ask_price2: float = float("nan")
        #: 卖2量
        self.ask_volume2: int = 0
        #: 买2价
        self.bid_price2: float = float("nan")
        #: 买2量
        self.bid_volume2: int = 0
        #: 卖3价
        self.ask_price3: float = float("nan")
        #: 卖3量
        self.ask_volume3: int = 0
        #: 买3价
        self.bid_price3: float = float("nan")
        #: 买3量
        self.bid_volume3: int = 0
        #: 卖4价
        self.ask_price4: float = float("nan")
        #: 卖4量
        self.ask_volume4: int = 0
        #: 买4价
        self.bid_price4: float = float("nan")
        #: 买4量
        self.bid_volume4: int = 0
        #: 卖5价
        self.ask_price5: float = float("nan")
        #: 卖5量
        self.ask_volume5: int = 0
        #: 买5价
        self.bid_price5: float = float("nan")
        #: 买5量
        self.bid_volume5: int = 0
        #: 当日成交量
        self.volume: int = 0
        #: 成交额
        self.amount: float = float("nan")
        #: 持仓量
        self.open_interest: int = 0


class Account(Entity):
    """ Account 是一个账户对象 """

    def __init__(self, api):
        self._api = api
        #: 币种
        self.currency: str = ""
        #: 昨日账户权益(不包含期权)
        self.pre_balance: float = float("nan")
        #: 静态权益 （静态权益 = 昨日结算的权益 + 今日入金 - 今日出金, 以服务器查询ctp后返回的金额为准）(不包含期权)
        self.static_balance: float = float("nan")
        #: 账户权益 （账户权益 = 动态权益 = 静态权益 + 平仓盈亏 + 持仓盈亏 - 手续费 + 权利金 + 期权市值）
        self.balance: float = float("nan")
        #: 可用资金（可用资金 = 账户权益 - 冻结保证金 - 保证金 - 冻结权利金 - 冻结手续费 - 期权市值）
        self.available: float = float("nan")
        #: 期货公司返回的balance（ctp_balance = 静态权益 + 平仓盈亏 + 持仓盈亏 - 手续费 + 权利金）
        self.ctp_balance: float = float("nan")
        #: 期货公司返回的available（ctp_available = ctp_balance - 保证金 - 冻结保证金 - 冻结手续费 - 冻结权利金）
        self.ctp_available: float = float("nan")
        #: 浮动盈亏
        self.float_profit: float = float("nan")
        #: 持仓盈亏
        self.position_profit: float = float("nan")
        #: 本交易日内平仓盈亏
        self.close_profit: float = float("nan")
        #: 冻结保证金
        self.frozen_margin: float = float("nan")
        #: 保证金占用
        self.margin: float = float("nan")
        #: 冻结手续费
        self.frozen_commission: float = float("nan")
        #: 本交易日内交纳的手续费
        self.commission: float = float("nan")
        #: 冻结权利金
        self.frozen_premium: float = float("nan")
        #: 本交易日内收入-交纳的权利金
        self.premium: float = float("nan")
        #: 本交易日内的入金金额
        self.deposit: float = float("nan")
        #: 本交易日内的出金金额
        self.withdraw: float = float("nan")
        #: 风险度（风险度 = 保证金 / 账户权益）
        self.risk_ratio: float = float("nan")
        #: 期权市值
        self.market_value: float = float("nan")


class Position(Entity):
    """ Position 是一个持仓对象 """

    def __init__(self, api):
        self._api = api
        #: 交易所
        self.exchange_id: str = ""
        #: 交易所内的合约代码
        self.instrument_id: str = ""
        #: 多头老仓手数
        self.pos_long_his: int = 0
        #: 多头今仓手数
        self.pos_long_today: int = 0
        #: 空头老仓手数
        self.pos_short_his: int = 0
        #: 空头今仓手数
        self.pos_short_today: int = 0
        #: 期货公司查询的多头今仓手数 (不推荐, 推荐使用pos_long_today)
        self.volume_long_today: int = 0
        #: 期货公司查询的多头老仓手数 (不推荐, 推荐使用pos_long_his)
        self.volume_long_his: int = 0
        #: 期货公司查询的多头手数 (不推荐, 推荐使用pos_long)
        self.volume_long: int = 0
        #: 期货公司查询的多头今仓冻结 (不推荐)
        self.volume_long_frozen_today: int = 0
        #: 期货公司查询的多头老仓冻结 (不推荐)
        self.volume_long_frozen_his: int = 0
        #: 期货公司查询的多头持仓冻结 (不推荐)
        self.volume_long_frozen: int = 0
        #: 期货公司查询的空头今仓手数 (不推荐, 推荐使用pos_short_today)
        self.volume_short_today: int = 0
        #: 期货公司查询的空头老仓手数 (不推荐, 推荐使用pos_short_his)
        self.volume_short_his: int = 0
        #: 期货公司查询的空头手数 (不推荐, 推荐使用pos_short)
        self.volume_short: int = 0
        #: 期货公司查询的空头今仓冻结 (不推荐)
        self.volume_short_frozen_today: int = 0
        #: 期货公司查询的空头老仓冻结 (不推荐)
        self.volume_short_frozen_his: int = 0
        #: 期货公司查询的空头持仓冻结 (不推荐)
        self.volume_short_frozen: int = 0
        #: 多头开仓均价,以开仓价来统计
        self.open_price_long: float = float("nan")
        #: 空头开仓均价,以开仓价来统计
        self.open_price_short: float = float("nan")
        #: 多头开仓成本,为开仓价乘以手数
        self.open_cost_long: float = float("nan")
        #: 空头开仓成本,为开仓价乘以手数
        self.open_cost_short: float = float("nan")
        #: 多头持仓均价,为多头持仓成本除以多头数量
        self.position_price_long: float = float("nan")
        #: 空头持仓均价,为空头持仓成本除以空头数量
        self.position_price_short: float = float("nan")
        #: 多头持仓成本,为今仓的开仓价乘以手数加上昨仓的昨结算价乘以手数的和
        self.position_cost_long: float = float("nan")
        #: 空头持仓成本,为今仓的开仓价乘以手数加上昨仓的昨结算价乘以手数的和
        self.position_cost_short: float = float("nan")
        #: 多头浮动盈亏
        self.float_profit_long: float = float("nan")
        #: 空头浮动盈亏
        self.float_profit_short: float = float("nan")
        #: 浮动盈亏 （浮动盈亏: 相对于开仓价的盈亏）
        self.float_profit: float = float("nan")
        #: 多头持仓盈亏
        self.position_profit_long: float = float("nan")
        #: 空头持仓盈亏
        self.position_profit_short: float = float("nan")
        #: 持仓盈亏 （持仓盈亏: 相对于上一交易日结算价的盈亏），期权持仓盈亏为 0
        self.position_profit: float = float("nan")
        #: 多头占用保证金
        self.margin_long: float = float("nan")
        #: 空头占用保证金
        self.margin_short: float = float("nan")
        #: 占用保证金
        self.margin: float = float("nan")
        #: 期权权利方市值(始终 >= 0)
        self.market_value_long: float = float("nan")
        #: 期权义务方市值(始终 <= 0)
        self.market_value_short: float = float("nan")
        #: 期权市值
        self.market_value: float = float("nan")
        #: 净持仓手数, ==0表示无持仓或多空持仓手数相等. <0表示空头持仓大于多头持仓, >0表示多头持仓大于空头持仓
        self.pos: int = 0
        #: 多头持仓手数, ==0表示无多头持仓. >0表示多头持仓手数
        self.pos_long: int = 0
        #: 空头持仓手数, ==0表示无空头持仓. >0表示空头持仓手数
        self.pos_short: int = 0

    @property
    def orders(self):
        """
        与此持仓相关的且目前委托单状态为ALIVE的开仓/平仓挂单

        :return: dict, 其中每个元素的key为委托单ID, value为 :py:class:`~tqsdk.objs.Order`
        """
        tdict = _get_obj(self._api._data, ["trade", self._path[1], "orders"])
        fts = {order_id: order for order_id, order in tdict.items() if (not order_id.startswith(
            "_")) and order.instrument_id == self.instrument_id and order.exchange_id == self.exchange_id and order.status == "ALIVE"}
        return fts


class Order(Entity):
    """ Order 是一个委托单对象 """

    def __init__(self, api):
        self._api = api
        #: 委托单ID, 对于一个用户的所有委托单，这个ID都是不重复的
        self.order_id: str = ""
        #: 交易所单号
        self.exchange_order_id: str = ""
        #: 交易所
        self.exchange_id: str = ""
        #: 交易所内的合约代码
        self.instrument_id: str = ""
        #: 下单方向, BUY=买, SELL=卖
        self.direction: str = ""
        #: 开平标志, OPEN=开仓, CLOSE=平仓, CLOSETODAY=平今
        self.offset: str = ""
        #: 总报单手数
        self.volume_orign: int = 0
        #: 未成交手数
        self.volume_left: int = 0
        #: 委托价格, 仅当 price_type = LIMIT 时有效
        self.limit_price: float = float("nan")
        #: 价格类型, ANY=市价, LIMIT=限价
        self.price_type: str = ""
        #: 手数条件, ANY=任何数量, MIN=最小数量, ALL=全部数量
        self.volume_condition: str = ""
        #: 时间条件, IOC=立即完成，否则撤销, GFS=本节有效, GFD=当日有效, GTC=撤销前有效, GFA=集合竞价有效
        self.time_condition: str = ""
        #: 下单时间，自unix epoch(1970-01-01 00:00:00 GMT)以来的纳秒数.
        self.insert_date_time: int = 0
        #: 委托单状态信息
        self.last_msg: str = ""
        #: 委托单状态, ALIVE=有效, FINISHED=已完
        self.status: str = ""
        #: 委托单是否确定已死亡（以后一定不会再产生成交）(注意，False 不代表委托单还存活，有可能交易所回来的信息还在路上或者丢掉了)
        self.is_dead: bool = None
        #: 委托单是否确定已报入交易所并等待成交 (注意，返回 False 不代表确定未报入交易所，有可能交易所回来的信息还在路上或者丢掉了)
        self.is_online: bool = None
        #: 委托单是否确定是错单（即下单失败，一定不会有成交）(注意，返回 False 不代表确定不是错单，有可能交易所回来的信息还在路上或者丢掉了)
        self.is_error: bool = None
        #: 平均成交价
        self.trade_price: float = float('nan')

        self._this_session = False

    @property
    def trade_records(self):
        """
        成交记录

        :return: dict, 其中每个元素的key为成交ID, value为 :py:class:`~tqsdk.objs.Trade`
        """
        tdict = _get_obj(self._api._data, ["trade", self._path[1], "trades"])
        fts = {trade_id: trade for trade_id, trade in tdict.items() if
               (not trade_id.startswith("_")) and trade.order_id == self.order_id}
        return fts


class Trade(Entity):
    """ Trade 是一个成交对象 """

    def __init__(self, api):
        self._api = api
        #: 委托单ID, 对于一个用户的所有委托单，这个ID都是不重复的
        self.order_id: str = ""
        #: 成交ID, 对于一个用户的所有成交，这个ID都是不重复的
        self.trade_id: str = ""
        #: 交易所成交编号
        self.exchange_trade_id: str = ""
        #: 交易所
        self.exchange_id: str = ""
        #: 交易所内的合约代码
        self.instrument_id: str = ""
        #: 下单方向, BUY=买, SELL=卖
        self.direction: str = ""
        #: 开平标志, OPEN=开仓, CLOSE=平仓, CLOSETODAY=平今
        self.offset: str = ""
        #: 成交价格
        self.price: float = float("nan")
        #: 成交手数
        self.volume: int = 0
        #: 成交时间，自unix epoch(1970-01-01 00:00:00 GMT)以来的纳秒数
        self.trade_date_time: int = 0


class RiskManagementRule(Entity):

    def __init__(self, api):
        self._api = api
        self.user_id = ""
        self.exchange_id = ""
        self.enable = False
        self.self_trade = SelfTradeRule(self._api)
        self.frequent_cancellation = FrequentCancellationRule(self._api)
        self.trade_position_ratio = TradePositionRatioRule(self._api)

    def _instance_entity(self, path):
        super(RiskManagementRule, self)._instance_entity(path)
        self.self_trade = copy.copy(self.self_trade)
        self.self_trade._instance_entity(path + ["self_trade"])
        self.frequent_cancellation = copy.copy(self.frequent_cancellation)
        self.frequent_cancellation._instance_entity(path + ["frequent_cancellation"])
        self.trade_position_ratio = copy.copy(self.trade_position_ratio)
        self.trade_position_ratio._instance_entity(path + ["trade_position_ratio"])


class SelfTradeRule(Entity):
    """自成交风控规则"""

    def __init__(self, api):
        self._api = api
        #: 最大自成交次数限制
        self.count_limit = 0

    def __repr__(self):
        return json.dumps({"count_limit": self.count_limit})


class FrequentCancellationRule(Entity):
    """频繁报撤单风控规则"""

    def __init__(self, api):
        self._api = api
        #: 频繁报撤单起算报单次数
        self.insert_order_count_limit = 0
        #: 频繁报撤单起算撤单次数
        self.cancel_order_count_limit = 0
        #: 频繁报撤单撤单比例限额,为百分比
        self.cancel_order_percent_limit = float("nan")

    def __repr__(self):
        return json.dumps({
            "insert_order_count_limit": self.insert_order_count_limit,
            "cancel_order_count_limit": self.cancel_order_count_limit,
            "cancel_order_percent_limit": self.cancel_order_percent_limit
        })


class TradePositionRatioRule(Entity):
    """成交持仓比风控规则"""

    def __init__(self, api):
        self._api = api
        #: 成交持仓比起算成交手数
        self.trade_units_limit = 0
        #: 成交持仓比例限额,为百分比
        self.trade_position_ratio_limit = float("nan")

    def __repr__(self):
        return json.dumps({
            "trade_units_limit": self.trade_units_limit,
            "trade_position_ratio_limit": self.trade_position_ratio_limit
        })


class RiskManagementData(Entity):

    def __init__(self, api):
        self._api = api
        #: 用户ID
        self.user_id = ""
        #: 交易所ID
        self.exchange_id = ""
        #: 合约ID
        self.instrument_id = ""
        #: 自成交情况
        self.self_trade = SelfTrade(self._api)
        #: 频繁报撤单情况
        self.frequent_cancellation = FrequentCancellation(self._api)
        #: 成交持仓比情况
        self.trade_position_ratio = TradePositionRatio(self._api)

    def _instance_entity(self, path):
        super(RiskManagementData, self)._instance_entity(path)
        self.self_trade = copy.copy(self.self_trade)
        self.self_trade._instance_entity(path + ["self_trade"])
        self.frequent_cancellation = copy.copy(self.frequent_cancellation)
        self.frequent_cancellation._instance_entity(path + ["frequent_cancellation"])
        self.trade_position_ratio = copy.copy(self.trade_position_ratio)
        self.trade_position_ratio._instance_entity(path + ["trade_position_ratio"])


class SelfTrade(Entity):
    """自成交情况"""

    def __init__(self, api):
        self._api = api
        #: 当前最高买价
        self.highest_buy_price = float("nan")
        #: 当前最低卖价
        self.lowest_sell_price = float("nan")
        #: 当天已经发生的自成交次数
        self.self_trade_count = 0
        #: 当天由于自成交而被拒的报单次数
        self.rejected_count = 0

    def __repr__(self):
        return json.dumps({
            "highest_buy_price": self.highest_buy_price,
            "lowest_sell_price": self.lowest_sell_price,
            "self_trade_count": self.self_trade_count,
            "rejected_count": self.rejected_count
        })


class FrequentCancellation(Entity):
    """频繁报撤单情况"""

    def __init__(self, api):
        self._api = api
        #: 当天已经发生的报单次数
        self.insert_order_count = 0
        #: 当天已经发生的撤单次数
        self.cancel_order_count = 0
        #: 当天的撤单比例，为百分比
        self.cancel_order_percent = float("nan")
        #: 当天由于撤单比例超限而被拒的撤单次数
        self.rejected_count = 0

    def __repr__(self):
        return json.dumps({
            "insert_order_count": self.insert_order_count,
            "cancel_order_count": self.cancel_order_count,
            "cancel_order_percent": self.cancel_order_percent,
            "rejected_count": self.rejected_count
        })


class TradePositionRatio(Entity):
    """成交持仓比情况"""

    def __init__(self, api):
        self._api = api
        #: 当天已经发生的成交手数
        self.trade_units = 0
        #: 当前的净持仓手数, 正为多仓, 负为空仓
        self.net_position_units = 0
        #: 当前的成交持仓比, 为百分比
        self.trade_position_ratio = float("nan")
        #: 当天由于成交持仓比超限而被拒的报单次数
        self.rejected_count = 0

    def __repr__(self):
        return json.dumps({
            "trade_units": self.trade_units,
            "net_position_units": self.net_position_units,
            "trade_position_ratio": self.trade_position_ratio,
            "rejected_count": self.rejected_count
        })


class SecurityAccount(Entity):
    """ SecurityAccount 是一个股票账户对象"""

    def __init__(self, api):
        self._api = api
        #: 用户客户号
        self.user_id: str = ""
        #: 币种, CNY=人民币, USD=美元, HKD=港币
        self.currency: str = "CNY"
        #: 当前市值
        self.market_value: float = float("nan")
        #: 当前资产 = 当前市值 + 可用金额 + 委托冻结金额 + 委托冻结费用
        self.asset: float = float("nan")
        #: 期初资产
        self.asset_his: float = float("nan")
        #: 当前可用余额 = 期初可用 + 当日入金 - 当日出金 + 当日分红金额 - 当日买入金额 - 当日买入费用 + 当日卖出金额 - 当日卖出费用 - 委托冻结金额 - 委托冻结手续费
        self.available: float = float("nan")
        #: 期初可用余额
        self.available_his: float = float("nan")
        #: 当前买入成本
        self.cost: float = float("nan")
        #: 当前可取余额 = MAX(期初余额 + 当日入金 – 当日出金 + MIN(0，当日卖出释放资金 - 当日买入占用资金 - 委托冻结金额), 0)
        self.drawable: float = float("nan")
        #: 本交易日内累计入金金额
        self.deposit: float = float("nan")
        #: 本交易日内累计出金金额
        self.withdraw: float = float("nan")
        #: 当前交易冻结金额（不含费用）= sum(order.volume_orign * order.limit_price)
        self.buy_frozen_balance: float = float("nan")
        #: 当前交易冻结费用 = sum(order.frozen_fee)
        self.buy_frozen_fee: float = float("nan")
        #: 当日买入占用资金（不含费用）
        self.buy_balance_today: float = float("nan")
        #: 当日买入累计费用
        self.buy_fee_today: float = float("nan")
        #: 当日卖出释放资金
        self.sell_balance_today: float = float("nan")
        #: 当日卖出累计费用
        self.sell_fee_today: float = float("nan")
        #: 当日持仓盈亏 = 当前市值 - 当前买入成本
        self.hold_profit: float = float("nan")
        #: 当日浮动盈亏 = SUM(持仓当日浮动盈亏)
        self.float_profit_today: float = float("nan")
        #: 当日实现盈亏 = SUM(持仓当日实现盈亏)
        self.real_profit_today: float = float("nan")
        #: 当日盈亏 = 当日浮动盈亏 + 当日实现盈亏
        self.profit_today: float = float("nan")
        #: 当日盈亏比 = 当日盈亏 / (当前买入成本 if 当前买入成本 > 0 else 期初资产)
        self.profit_rate_today: float = float("nan")
        #: 当日分红金额
        self.dividend_balance_today: float = float("nan")


class SecurityPosition(Entity):
    """ SecurityPosition 是一个股票账户持仓对象 """

    def __init__(self, api):
        self._api = api
        #: 用户客户号
        self.user_id: str = ""
        #: 交易所
        self.exchange_id: str = ""
        #: 证券代码
        self.instrument_id: str = ""
        #: 建仓日期
        self.create_date: str = ""
        #: 当前成本 = 昨买入成本 + 今买金额 + 今买费用 - 今卖数量 × (昨买入成本 / 昨持仓数量)
        self.cost: float = float("nan")
        #: 期初成本
        self.cost_his: float = float("nan")
        #: 今持仓数量 = 昨持仓数量 + 今买数量 - 今卖数量 + 送股数量
        self.volume: int = 0
        #: 昨持仓数量
        self.volume_his: int = 0
        #: 最新价
        self.last_price: float = float("nan")
        #: 当日累计买入持仓
        self.buy_volume_today: int = 0
        #: 当日累计买入金额 (不包括费用)
        self.buy_balance_today: float = float("nan")
        #: 当日累计买入费用
        self.buy_fee_today: float = float("nan")
        #: 当日累计卖出持仓
        self.sell_volume_today: int = 0
        #: 当日累计卖出金额(不包括费用)
        self.sell_balance_today: float = float("nan")
        #: 当日累计卖出费用
        self.sell_fee_today: float = float("nan")
        #: 期初累计买入持仓
        self.buy_volume_his: int = 0
        #: 期初累计买入金额
        self.buy_balance_his: float = float("nan")
        #: 期初累计买入费用
        self.buy_fee_his: float = float("nan")
        #: 期初累计卖出持仓
        self.sell_volume_his: int = 0
        #: 期初累计卖出金额
        self.sell_balance_his: float = float("nan")
        #: 期初累计卖出费用
        self.sell_fee_his: float = float("nan")
        #: 今送股数量
        self.shared_volume_today: float = float("nan")
        #: 今分红金额
        self.devidend_balance_today: float = float("nan")
        #: 当前市值 = 持仓数量 × 行情最新价
        self.market_value: float = float("nan")
        #: 期初市值
        self.market_value_his: float = float("nan")
        #: 当日浮动盈亏 = (昨持仓数量 - 今卖数量) * (最新价 - 昨收盘价) + (今持仓数量 - (昨持仓数量 - 今卖数量)) * (最新价 - 买入均价)
        self.float_profit_today: float = float("nan")
        #: 当日实现盈亏 = 今卖数量 * (最新价 - 昨收盘价) - 今卖费用 + 今派息金额
        self.real_profit_today: float = float("nan")
        #: 期初实现盈亏
        self.real_profit_his: float = float("nan")
        #: 当日盈亏 = 当日浮动盈亏 + 当日实现盈亏
        self.profit_today: float = float("nan")
        #: 当日收益率
        self.profit_rate_today: float = float("nan")
        #: 当日持仓盈亏 = 当前持仓市值 – 当前买入成本
        self.hold_profit: float = float("nan")
        #: 累计实现盈亏 += 当日实现盈亏（成本）
        self.real_profit_total: float = float("nan")
        #: 总盈亏 = 累计实现盈亏 + 持仓盈亏
        self.profit_total: float = float("nan")
        #: 累计收益率 = 总盈亏  / (期初成本 if 当前成本==0 else 当前成本)
        self.profit_rate_total: float = float("nan")

    @property
    def orders(self):
        tdict = _get_obj(self._api._data, ["trade", self._path[1], "orders"])
        fts = {order_id: order for order_id, order in tdict.items() if (not order_id.startswith(
            "_")) and order.instrument_id == self.instrument_id and order.exchange_id == self.exchange_id and order.status == "ALIVE"}
        return fts


class SecurityOrder(Entity):
    """ SecurityOrder 是一个股票账户委托单对象 """

    def __init__(self, api):
        self._api = api
        #: 用户客户号
        self.user_id: str = ""
        #: 订单号,要求客户端保证其在一个交易日内的唯一性
        self.order_id: str = ""
        #: 交易所委托合同编号
        self.exchange_order_id: str = ""
        #: 交易所代码
        self.exchange_id: str = ""
        #: 证券代码
        self.instrument_id: str = ""
        #: 下单方向, BUY=买, SELL=卖
        self.direction: str = ""
        #: 委托股数（A 股委托数量必须为 100 倍数；科创板股票必须为 200 倍数；零股卖出: 由于部分成交或分红导致持仓数量小于 100 股时，该部分持仓可一次性卖出，数量不为 100 或 200 倍数）
        self.volume_orign: int = 0
        #: 剩余股数
        self.volume_left: int = 0
        #: 报单价格类型，LIMIT=限价单，ANY=市价单
        self.price_type: str = ""
        #: 报单委托价格
        self.limit_price: float = float("nan")
        #: 冻结费用
        self.frozen_fee: float = float("nan")
        #: 委托时间，自unix epoch(1970-01-01 00:00:00 GMT)以来的纳秒数.
        self.insert_date_time: int = 0
        #: 委托单当前状态
        self.status: str = ""
        #: 委托单状态信息
        self.last_msg: str = ""

    @property
    def trade_records(self):
        """
        成交记录

        :return: dict, 其中每个元素的key为成交ID, value为 :py:class:`~tqsdk.objs.Trade`
        """
        tdict = _get_obj(self._api._data, ["trade", self._path[1], "trades"])
        fts = {trade_id: trade for trade_id, trade in tdict.items() if
               (not trade_id.startswith("_")) and trade.order_id == self.order_id}
        return fts


class SecurityTrade(Entity):
    """ SecurityTrade 是一个股票账户成交对象 """

    def __init__(self, api):
        self._api = api
        #: 用户客户号
        self.user_id: str = ""
        #: 成交编号
        self.trade_id: str = ""
        #: 交易所代码
        self.exchange_id: str = ""
        #: 证券代码
        self.instrument_id: str = ""
        #: 委托单编号
        self.order_id: str = ""
        #: 交易所订单编号
        self.exchange_order_id: str = ""
        #: 下单方向, BUY=买, SELL=卖，SHARED=送股，DEVIDEND=分红
        self.direction: str = ""
        #: 成交数量或者送股数量
        self.volume: int = 0
        #: 成交价格
        self.price: float = float("nan")
        #: 成交发生金额或分红金额
        self.balance: float = float("nan")
        #: 费用
        self.fee: float = float("nan")
        #: 成交时间，自unix epoch(1970-01-01 00:00:00 GMT)以来的纳秒数.
        self.trade_date_time: int = 0



#!usr/bin/env python3
# -*- coding:utf-8 -*-
__author__ = 'yanqiong'

import asyncio
import os
import socket
import sys
from datetime import datetime
from urllib.parse import urlparse

import numpy as np
import simplejson
from aiohttp import web
from tqsdk.tradeable.sim.basesim import BaseSim

from tqsdk.auth import TqAuth
from tqsdk.backtest import TqBacktest, TqReplay
from tqsdk.channel import TqChan
from tqsdk.datetime import _get_trading_day_start_time, _datetime_to_timestamp_nano
from tqsdk.diff import _simple_merge_diff
from tqsdk.tradeable import TqAccount, TqKq, TqSim


class TqWebHelper(object):

    def __init__(self, api):
        """初始化，检查参数"""
        self._api = api
        ip, port = TqWebHelper.parse_url(self._api._web_gui)
        self._http_server_host = ip if ip else "0.0.0.0"
        self._http_server_port = int(port) if port else 0
        args = TqWebHelper.parser_env_arguments()

        if args["_action"] == "run":
            # 运行模式下，账户参数冲突需要抛错，提示用户
            if args["_broker_id"] == "TQ_KQ":
                if not isinstance(self._api._account, TqKq) and not isinstance(self._api._account, TqSim):
                    raise Exception("策略代码与插件设置中的账户参数冲突。可尝试删去代码中的账户参数，以插件设置的账户参数运行。")
                self._api._account = TqKq()
            elif args["_broker_id"] and args["_account_id"] and args["_password"]:
                if isinstance(self._api._account, TqSim):
                    pass
                elif isinstance(self._api._account, TqAccount) and self._api._account._account_id == args["_account_id"] and self._api._account._broker_id == args["_broker_id"]:
                    pass
                else:
                    raise Exception("策略代码与插件设置中的账户参数冲突。可尝试删去代码中的账户参数，以插件设置的账户参数运行。")
                self._api._account = TqAccount(args["_broker_id"], args["_account_id"], args["_password"])
            else:
                self._api._account = TqSim(args["_init_balance"])
            self._api._backtest = None
            self._api._print(f"正在使用账户 {args['_broker_id']}, {args['_account_id']} 运行策略。")
        elif args["_action"] is not None:
            self._api._account = TqSim(args["_init_balance"])
            if args["_action"] == "backtest":
                self._api._backtest = TqBacktest(start_dt=datetime.strptime(args["_start_dt"], '%Y%m%d'),
                                                 end_dt=datetime.strptime(args["_end_dt"], '%Y%m%d'))
                self._api._print(f"当前回测区间 {args['_start_dt']} - {args['_end_dt']}。")
            elif args["_action"] == "replay":
                self._api._backtest = TqReplay(datetime.strptime(args["_replay_dt"], '%Y%m%d'))
                self._api._print(f"当前复盘日期 {args['_replay_dt']}。")
        if args["_auth"]:
            comma_index = args["_auth"].find(',')
            user_name, pwd = args["_auth"][:comma_index], args["_auth"][comma_index + 1:]
            if self._api._auth is not None and (user_name != self._api._auth._user_name or pwd != self._api._auth._password):
                raise Exception("策略代码与插件设置中的 auth 参数冲突。可尝试删去代码中的 auth 参数，以插件设置的参数运行。")
            self._api._auth = TqAuth(user_name, pwd)
        if args["_http_server_address"]:
            self._api._web_gui = True  # 命令行 _http_server_address, 一定打开 _web_gui
            ip, port = TqWebHelper.parse_url(args["_http_server_address"])
            self._http_server_host = ip if ip else "0.0.0.0"
            self._http_server_port = int(port) if port else 0

    async def _run(self, api_send_chan, api_recv_chan, web_send_chan, web_recv_chan):
        if not self._api._web_gui:
            # 没有开启 web_gui 功能
            _data_handler_without_web_task = self._api.create_task(
                self._data_handler_without_web(api_recv_chan, web_recv_chan))
            try:
                async for pack in api_send_chan:
                    # api 发送的包，过滤出 set_chart_data, 其余的原样转发
                    if pack['aid'] not in ['set_chart_data', 'set_report_data']:
                        await web_send_chan.send(pack)
            finally:
                _data_handler_without_web_task.cancel()
                await asyncio.gather(_data_handler_without_web_task, return_exceptions=True)
        else:
            self._web_dir = os.path.join(os.path.dirname(__file__), 'web')
            file_path = os.path.abspath(sys.argv[0])
            file_name = os.path.basename(file_path)
            # 初始化数据截面
            accounts_info = {
                #@todo: 避免针对账户类型的特殊处理, 统一按照通知的形式更新连接状态, 同时解决回测模式下没有行情连接状态的问题
                acc._account_key: {"td_url_status": True if isinstance(acc, BaseSim) else '-'}
                for acc in self._api._account._account_list
            }
            for acc in self._api._account._account_list:
                accounts_info[acc._account_key].update(acc._account_info)
            self._data = {
                "action": {
                    "mode": "replay" if isinstance(self._api._backtest, TqReplay) else "backtest" if isinstance(self._api._backtest, TqBacktest) else "run",
                    "md_url_status": '-',
                    "user_name": self._api._auth._user_name,
                    "file_path": file_path[0].upper() + file_path[1:],
                    "file_name": file_name,
                    "accounts": accounts_info
                },
                "trade": {},
                "subscribed": [],
                "draw_chart_datas": {},
                "snapshots": {}
            }
            self._order_symbols = set()
            self._diffs = []
            self._conn_diff_chans = set()
            _data_task = self._api.create_task(self._data_handler(api_recv_chan, web_recv_chan))
            _httpserver_task = self._api.create_task(self.link_httpserver())

            try:
                # api 发送的包，过滤出需要的包记录在 self._data
                async for pack in api_send_chan:
                    if pack['aid'] == 'set_chart_data':
                        # 发送的是绘图数据
                        diff_data = {}  # 存储 pack 中的 diff 数据的对象
                        for series_id, series in pack['datas'].items():
                            diff_data[series_id] = series
                        if diff_data != {}:
                            web_diff = {'draw_chart_datas': {}}
                            web_diff['draw_chart_datas'][pack['symbol']] = {}
                            web_diff['draw_chart_datas'][pack['symbol']][pack['dur_nano']] = diff_data
                            _simple_merge_diff(self._data, web_diff)
                            for chan in self._conn_diff_chans:
                                self.send_to_conn_chan(chan, [web_diff])
                    elif pack['aid'] == 'set_report_data':
                        # 发送的是绘图报告数据
                        web_diff = {'draw_report_datas': {}}
                        for data in pack['report_datas']:
                            _simple_merge_diff(web_diff['draw_report_datas'], data)
                        _simple_merge_diff(self._data, web_diff)
                        for chan in self._conn_diff_chans:
                            self.send_to_conn_chan(chan, [web_diff])
                    else:
                        if pack["aid"] == "insert_order":
                            self._order_symbols.add(pack["exchange_id"] + "." + pack["instrument_id"])
                        if pack['aid'] == 'subscribe_quote' or pack["aid"] == "set_chart" or pack["aid"] == 'insert_order':
                            web_diff = {'subscribed': []}
                            for item in self._api._requests["klines"].keys():
                                web_diff['subscribed'].append({"symbol": item[0], "dur_nano": item[1] * 1000000000})
                            for item in self._api._requests["ticks"].keys():
                                web_diff['subscribed'].append({"symbol": item[0], "dur_nano": 0})
                            for symbol in self._api._requests["quotes"]:
                                web_diff['subscribed'].append({"symbol": symbol})
                            for symbol in self._order_symbols:
                                web_diff['subscribed'].append({"symbol": symbol})
                            if web_diff['subscribed'] != self._data['subscribed']:
                                self._data['subscribed'] = web_diff['subscribed']
                            for chan in self._conn_diff_chans:
                                self.send_to_conn_chan(chan, [web_diff])
                        # 发送的转发给上游
                        await web_send_chan.send(pack)
            finally:
                _data_task.cancel()
                _httpserver_task.cancel()
                await asyncio.gather(_data_task, _httpserver_task, return_exceptions=True)

    async def _data_handler_without_web(self, api_recv_chan, web_recv_chan):
        # 没有 web_gui, 接受全部数据转发给下游 api_recv_chan
        async for pack in web_recv_chan:
            await api_recv_chan.send(pack)

    async def _data_handler(self, api_recv_chan, web_recv_chan):
        async for pack in web_recv_chan:
            if pack['aid'] == 'rtn_data':
                web_diffs = []
                account_changed = False
                for d in pack['data']:
                    # 把 d 处理成需要的数据
                    # 处理 trade
                    trade = d.get("trade")
                    if trade is not None:
                        _simple_merge_diff(self._data["trade"], trade)
                        web_diffs.append({"trade": trade})
                        # 账户是否有变化
                        # todo: 这里本来使用类似 is_changing 的做法，判断 diffs 中是否有 static_balance 字段；由于 _simple_merge_diff 去掉了默认参数 reduce_diff
                        #  现在修改为手动比较 diffs 中的 static_balance 和 self._data 中的 static_balance 是否一样
                        account_key = self._api._account._get_account_key(account=None)
                        current_static_balance = self._data["trade"].get(account_key, {}).get("accounts", {}).get("CNY", {}).get('static_balance')
                        diff_static_balance = d.get("trade", {}).get(account_key, {}).get("accounts", {}).get("CNY", {}).get('static_balance', None)
                        static_balance_changed = diff_static_balance is not None and current_static_balance != diff_static_balance
                        trades_changed = d.get("trade", {}).get(account_key, {}).get("trades", {})
                        orders_changed = d.get("trade", {}).get(account_key, {}).get("orders", {})
                        if static_balance_changed is True or trades_changed != {} or orders_changed != {}:
                            account_changed = True
                    # 处理 backtest replay
                    if d.get("_tqsdk_backtest") or d.get("_tqsdk_replay"):
                        _simple_merge_diff(self._data, d)
                        web_diffs.append(d)
                    # 处理通知，行情和交易连接的状态
                    notify_diffs = self._notify_handler(d.get("notify", {}))
                    for diff in notify_diffs:
                        _simple_merge_diff(self._data, diff)
                    web_diffs.extend(notify_diffs)
                if account_changed:
                    dt, snapshot = self.get_snapshot()
                    _snapshots = {"snapshots": {}}
                    _snapshots["snapshots"][dt] = snapshot
                    web_diffs.append(_snapshots)
                    _simple_merge_diff(self._data, _snapshots)
                for chan in self._conn_diff_chans:
                    self.send_to_conn_chan(chan, web_diffs)
            # 接收的数据转发给下游 api
            await api_recv_chan.send(pack)

    def _notify_handler(self, notifies):
        """将连接状态的通知转成 diff 协议"""
        diffs = []
        for _, notify in notifies.items():
            if notify["code"] == 2019112901 or notify["code"] == 2019112902:
                url_status = True  # 连接建立的通知 第一次建立 或者 重连建立
            elif notify["code"] == 2019112911:
                url_status = False  # 连接断开的通知
            else:
                continue
            if notify["url"] == self._api._md_url:
                diffs.append({
                    "action": {"md_url_status": url_status}
                })
            elif notify["conn_id"] in self._api._account._map_conn_id:
                acc = self._api._account._map_conn_id[notify["conn_id"]]
                diffs.append({
                    "action": {
                        "accounts": {acc._account_key: {"td_url_status": url_status}}
                    }
                })
        return diffs

    def send_to_conn_chan(self, chan, diffs):
        last_diff = chan.recv_latest({})
        for d in diffs:
            _simple_merge_diff(last_diff, d)
        if last_diff != {}:
            chan.send_nowait(last_diff)

    def dt_func (self):
        # 回测和复盘模式，用 _api._account 一定是 TqSim, 使用 TqSim _get_current_timestamp() 提供的时间
        # todo: 使用 TqSim.EPOCH
        if self._data["action"]["mode"] == "backtest":
            return self._data['_tqsdk_backtest']['current_dt']
        elif self._data["action"]["mode"] == "replay":
            tqsim_current_timestamp = self._api._account._account_list[0]._get_current_timestamp()
            if tqsim_current_timestamp == 631123200000000000:
                # 未收到任何行情, TqSim 时间没有更新
                return _get_trading_day_start_time(self._data['_tqsdk_replay']['replay_dt'])
            else:
                return tqsim_current_timestamp
        else:
            return _datetime_to_timestamp_nano(datetime.now())

    def get_snapshot(self):
        account = self._data.get("trade", {}).get(self._api._account._get_account_key(account=None), {}).get("accounts", {}).get("CNY", {})
        positions = self._data.get("trade", {}).get(self._api._account._get_account_key(account=None), {}).get("positions", {})
        dt = self.dt_func()
        return dt, {
            'accounts': {'CNY': {k: v for k, v in account.items() if not k.startswith("_")}},
            'positions': {k: {pk: pv for pk, pv in v.items() if not pk.startswith("_")} for k, v in
                          positions.items() if
                          not k.startswith("_")}
        }

    def get_send_msg(self, data=None):
        return simplejson.dumps({
            'aid': 'rtn_data',
            'data': [self._data if data is None else data]
        }, ignore_nan=True, default=TqWebHelper._convert)

    async def connection_handler(self, request):
        ws = web.WebSocketResponse()
        await ws.prepare(request)
        send_msg = self.get_send_msg(self._data)
        await ws.send_str(send_msg)
        conn_chan = TqChan(self._api, last_only=True)
        self._conn_diff_chans.add(conn_chan)
        try:
            async for msg in ws:
                pack = simplejson.loads(msg.data)
                if pack["aid"] == 'peek_message':
                    last_diff = await conn_chan.recv()
                    send_msg = self.get_send_msg(last_diff)
                    await ws.send_str(send_msg)
        except Exception as e:
            await conn_chan.close()
            self._conn_diff_chans.remove(conn_chan)

    async def link_httpserver(self):
        try:
            url_response = {
                "ins_url": self._api._ins_url,
                "md_url": self._api._md_url,
            }
            # TODO：在复盘模式下发送 replay_dt 给 web 端，服务器改完后可以去掉
            if isinstance(self._api._backtest, TqReplay):
                url_response["replay_dt"] = _datetime_to_timestamp_nano(datetime.combine(self._api._backtest._replay_dt, datetime.min.time()))
            app = web.Application()
            app.router.add_get(path='/url', handler=lambda request: TqWebHelper.httpserver_url_handler(url_response))
            app.router.add_get(path='/', handler=self.httpserver_index_handler)
            app.router.add_get(path='/index.html', handler=self.httpserver_index_handler)
            app.router.add_static('/web', self._web_dir, show_index=True)
            app.add_routes([web.get('/ws', self.connection_handler)])
            runner = web.AppRunner(app)
            await runner.setup()
            server_socket = socket.socket()
            if sys.platform.startswith("linux") or sys.platform.startswith("darwin"):
                server_socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
            server_socket.bind((self._http_server_host, self._http_server_port))
            address = server_socket.getsockname()
            site = web.SockSite(runner, server_socket)
            await site.start()
            ip = "127.0.0.1" if address[0] == "0.0.0.0" else address[0]
            self._api._print("您可以访问 http://{ip}:{port} 查看策略绘制出的 K 线图形。".format(ip=ip, port=address[1]))
            await asyncio.sleep(100000000000)
        finally:
            await runner.shutdown()
            await runner.cleanup()

    def httpserver_index_handler(self, request):
        return web.FileResponse(self._web_dir + '/index.html')

    @staticmethod
    def _convert(o):
        """对于 numpy 类型的数据，返回可以序列化的值"""
        if isinstance(o, np.generic):
            return o.item()
        raise TypeError

    @staticmethod
    def parse_url(url):
        if isinstance(url, str):
            parse_result = urlparse(url, scheme='')
            addr = parse_result.netloc if parse_result.scheme == "http" else url
            return addr.split(':')
        else:
            return '0.0.0.0', '0'

    @staticmethod
    def httpserver_url_handler(response):
        return web.json_response(response)

    @staticmethod
    def parser_env_arguments():
        action = {
            "_action": os.getenv("TQ_ACTION"),
            "_http_server_address": os.getenv("TQ_HTTP_SERVER_ADDRESS"),
            "_auth": os.getenv("TQ_AUTH")
        }
        try:
            action["_init_balance"] = 10000000.0 if os.getenv("TQ_INIT_BALANCE") is None else float(
                os.getenv("TQ_INIT_BALANCE"))
        except ValueError:
            action["_init_balance"] = 10000000.0
        if action["_action"] == "run":
            action["_broker_id"] = os.getenv("TQ_BROKER_ID")
            action["_account_id"] = os.getenv("TQ_ACCOUNT_ID")
            action["_password"] = os.getenv("TQ_PASSWORD")
            if not action["_broker_id"]:
                action["_action"] = None
        elif action["_action"] == "backtest":
            action["_start_dt"] = os.getenv("TQ_START_DT")
            action["_end_dt"] = os.getenv("TQ_END_DT")
            if not action["_start_dt"] or not action["_end_dt"]:
                action["_action"] = None
        elif action["_action"] == "replay":
            action["_replay_dt"] = os.getenv("TQ_REPLAY_DT")
            if not action["_replay_dt"]:
                action["_action"] = None
        return action



#!usr/bin/env python3
# -*- coding:utf-8 -*-

__author__ = 'mayanqiong'

from tqsdk.tradeable.mixin import FutureMixin
from tqsdk.datetime import _timestamp_nano_to_str
from tqsdk.diff import _get_obj
from tqsdk.objs import Quote
from tqsdk.report import TqReport
from tqsdk.tradeable.sim.basesim import BaseSim
from tqsdk.tradeable.sim.trade_future import SimTrade
from tqsdk.tradeable.sim.utils import _get_future_margin, _get_commission


class TqSim(BaseSim, FutureMixin):
    """
    天勤模拟交易类

    该类实现了一个本地的模拟账户，并且在内部完成撮合交易，在回测和复盘模式下，只能使用 TqSim 账户来交易。

    限价单要求报单价格达到或超过对手盘价格才能成交, 成交价为报单价格, 如果没有对手盘(涨跌停)则无法成交

    市价单使用对手盘价格成交, 如果没有对手盘(涨跌停)则自动撤单

    模拟交易不会有部分成交的情况, 要成交就是全部成交
    """

    def __init__(self, init_balance: float = 10000000.0, account_id: str = None) -> None:
        """
        Args:
            init_balance (float): [可选]初始资金, 默认为一千万

            account_id (str): [可选]帐号, 默认为 TQSIM

        Example::

            # 修改TqSim模拟帐号的初始资金为100000
            from tqsdk import TqApi, TqSim, TqAuth
            api = TqApi(TqSim(init_balance=100000), auth=TqAuth("快期账户", "账户密码"))

        """
        if float(init_balance) <= 0:
            raise Exception("初始资金(init_balance) %s 错误, 请检查 init_balance 是否填写正确" % (init_balance))
        super(TqSim, self).__init__(account_id="TQSIM" if account_id is None else account_id,
                                    init_balance=float(init_balance),
                                    trade_class=SimTrade)

    @property
    def _account_info(self):
        info = super(TqSim, self)._account_info
        info.update({
            "account_type": self._account_type
        })
        return info

    def set_commission(self, symbol: str, commission: float=float('nan')):
        """
        设置指定合约模拟交易的每手手续费。

        Args:
            symbol (str): 合约代码

            commission (float): 每手手续费

        Returns:
            float: 设置的每手手续费

        Example::

            from tqsdk import TqSim, TqApi, TqAuth

            sim = TqSim()
            api = TqApi(sim, auth=TqAuth("快期账户", "账户密码"))

            sim.set_commission("SHFE.cu2112", 50)

            print(sim.get_commission("SHFE.cu2112"))
        """
        if commission != commission:
            raise Exception("合约手续费不可以设置为 float('nan')")
        quote = _get_obj(self._data, ["quotes", symbol], Quote(self._api if hasattr(self, "_api") else None))
        quote["user_commission"] = commission
        if self._quote_tasks.get(symbol):
            self._quote_tasks[symbol]["quote_chan"].send_nowait({
                "quotes": {symbol: {"user_commission": commission}}
            })
        return commission

    def set_margin(self, symbol: str, margin: float=float('nan')):
        """
        设置指定合约模拟交易的每手保证金。

        Args:
            symbol (str): 合约代码 (只支持期货合约)

            margin (float): 每手保证金

        Returns:
            float: 设置的每手保证金

        Example::

            from tqsdk import TqSim, TqApi, TqAuth

            sim = TqSim()
            api = TqApi(sim, auth=TqAuth("快期账户", "账户密码"))

            sim.set_margin("SHFE.cu2112", 26000)

            print(sim.get_margin("SHFE.cu2112"))
        """
        if margin != margin:
            raise Exception("合约手续费不可以设置为 float('nan')")
        quote = _get_obj(self._data, ["quotes", symbol], Quote(self._api if hasattr(self, "_api") else None))
        quote["user_margin"] = margin
        if self._quote_tasks.get(symbol):
            self._quote_tasks[symbol]["quote_chan"].send_nowait({
                "quotes": {symbol: {"user_margin": margin}}
            })
            # 当用户设置保证金时，用户应该得到的效果是：
            # 在调用 sim.set_margin() 之后，立即调用 api.get_position(symbol)，得到的 margin 字段应该按照新设置的保证金调整过，而且中间没有收到过行情更新包
            # 以下代码可以保证这个效果，说明：
            # 1. 持仓已经调整过:
            #   sim_trade 中持仓的 future_margin 字段更新，margin 会同时调整，那么 api 中持仓的 future_margin 更新时，margin 一定也已经更新
            # 2. 中间没有收到过行情更新包:
            #   前提1：根据 diff 协议，sim 收到 peek_message 时，会将缓存的 diffs 发给用户，当缓存的 diffs 为空，会转发 peek_message；
            #   前提2：api.wait_update() 会等到所有 task 都执行到 pending 状态，然后发送 peek_message 给 sim
            #   当用户代码执行到 sim.set_margin()，立即向 quote_chan 中发送一个数据包，quote_task 就会到 ready 状态，此时调用 wait_update()，
            #   到所有 task 执行到 pending 状态时，sim 的 diffs 中有数据了，此时收到 api 发来 peek_message 不会转发给上游，用户会先收到 sim 本身的账户数据，
            #   在下一次 wait_update，sim 的 diffs 为空，才会收到行情数据
            # 在回测时，以下代码应该只经历一次 wait_update
            while margin != self.get_position(symbol).get("future_margin"):
                self._api.wait_update()
        return margin

    def get_margin(self, symbol: str):
        """
        获取指定合约模拟交易的每手保证金。

        Args:
            symbol (str): 合约代码

        Returns:
            float: 返回合约模拟交易的每手保证金

        Example::

            from tqsdk import TqSim, TqApi, TqAuth

            sim = TqSim()
            api = TqApi(sim, auth=TqAuth("快期账户", "账户密码"))

            quote = api.get_quote("SHFE.cu2112")
            print(sim.get_margin("SHFE.cu2112"))
        """
        return _get_future_margin(self._data.get("quotes", {}).get(symbol, {}))

    def get_commission(self, symbol: str):
        """
        获取指定合约模拟交易的每手手续费

        Args:
            symbol (str): 合约代码

        Returns:
            float: 返回合约模拟交易的每手手续费

        Example::

            from tqsdk import TqSim, TqApi, TqAuth

            sim = TqSim()
            api = TqApi(sim, auth=TqAuth("快期账户", "账户密码"))

            quote = api.get_quote("SHFE.cu2112")
            print(sim.get_commission("SHFE.cu2112"))
        """
        return _get_commission(self._data.get("quotes", {}).get(symbol, {}))

    def _handle_on_alive(self, msg, order):
        """
        在 order 状态变为 ALIVE 调用，屏幕输出信息，打印日志
        """
        symbol = f"{order['exchange_id']}.{order['instrument_id']}"
        self._api._print(
            f"模拟交易下单 {self._account_name}, {order['order_id']}: 时间: {_timestamp_nano_to_str(order['insert_date_time'])}, "
            f"合约: {symbol}, 开平: {order['offset']}, 方向: {order['direction']}, 手数: {order['volume_left']}, "
            f"价格: {order.get('limit_price', '市价')}")
        self._logger.debug(msg, order_id=order["order_id"], datetime=order["insert_date_time"],
                           symbol=symbol, offset=order["offset"], direction=order["direction"],
                           volume_left=order["volume_left"], limit_price=order.get("limit_price", "市价"))

    def _handle_on_finished(self, msg, order):
        """
        在 order 状态变为 FINISHED 调用，屏幕输出信息，打印日志
        """
        self._api._print(f"模拟交易委托单 {self._account_name}, {order['order_id']}: {order['last_msg']}")
        self._logger.debug(msg, order_id=order["order_id"], last_msg=order["last_msg"], status=order["status"],
                           volume_orign=order["volume_orign"], volume_left=order["volume_left"])

    def _report(self):
        if not self.trade_log:
            return
        date_keys = sorted(self.trade_log.keys())
        self._api._print(f"模拟交易成交记录, 账户: {self._account_name}")
        for d in date_keys:
            for t in self.trade_log[d]["trades"]:
                symbol = t["exchange_id"] + "." + t["instrument_id"]
                self._api._print(f"时间: {_timestamp_nano_to_str(t['trade_date_time'])}, 合约: {symbol}, "
                                 f"开平: {t['offset']}, 方向: {t['direction']}, 手数: {t['volume']}, 价格: {t['price']:.3f},"
                                 f"手续费: {t['commission']:.2f}")

        self._api._print(f"模拟交易账户资金, 账户: {self._account_name}")
        for d in date_keys:
            account = self.trade_log[d]["account"]
            self._api._print(
                f"日期: {d}, 账户权益: {account['balance']:.2f}, 可用资金: {account['available']:.2f}, "
                f"浮动盈亏: {account['float_profit']:.2f}, 持仓盈亏: {account['position_profit']:.2f}, "
                f"平仓盈亏: {account['close_profit']:.2f}, 市值: {account['market_value']:.2f}, "
                f"保证金: {account['margin']:.2f}, 手续费: {account['commission']:.2f}, "
                f"风险度: {account['risk_ratio'] * 100:.2f}%")

        # TqReport 模块计算交易统计信息
        report = TqReport(report_id=self._account_id, trade_log=self.trade_log, quotes=self._data['quotes'])
        self.tqsdk_stat = report.default_metrics
        self._api._print(
            f"胜率: {self.tqsdk_stat['winning_rate'] * 100:.2f}%, 盈亏额比例: {self.tqsdk_stat['profit_loss_ratio']:.2f}, "
            f"收益率: {self.tqsdk_stat['ror'] * 100:.2f}%, 年化收益率: {self.tqsdk_stat['annual_yield'] * 100:.2f}%, "
            f"最大回撤: {self.tqsdk_stat['max_drawdown'] * 100:.2f}%, 年化夏普率: {self.tqsdk_stat['sharpe_ratio']:.4f},"
            f"年化索提诺比率: {self.tqsdk_stat['sortino_ratio']:.4f}")

        # 回测情况下，在计算报告之后，还会发送绘制图表请求，
        # 这样处理，用户不要修改代码，就能够看到报告图表
        if self._tqsdk_backtest:
            self._api.draw_report(report.full())



#!/usr/bin/env python
#  -*- coding: utf-8 -*-
__author__ = 'limin'

'''
Dual Thrust策略 (难度：中级)
参考: https://www.shinnytech.com/blog/dual-thrust
注: 该示例策略仅用于功能示范, 实盘时请根据自己的策略/经验进行修改
'''

from tqsdk import TqApi, TqAuth, TargetPosTask

SYMBOL = "DCE.jd2011"  # 合约代码
NDAY = 5  # 天数
K1 = 0.2  # 上轨K值
K2 = 0.2  # 下轨K值

api = TqApi(auth=TqAuth("快期账户", "账户密码"))
print("策略开始运行")

quote = api.get_quote(SYMBOL)
klines = api.get_kline_serial(SYMBOL, 24 * 60 * 60)  # 86400使用日线
target_pos = TargetPosTask(api, SYMBOL)


def dual_thrust(quote, klines):
    current_open = klines.iloc[-1]["open"]
    HH = max(klines.high.iloc[-NDAY - 1:-1])  # N日最高价的最高价
    HC = max(klines.close.iloc[-NDAY - 1:-1])  # N日收盘价的最高价
    LC = min(klines.close.iloc[-NDAY - 1:-1])  # N日收盘价的最低价
    LL = min(klines.low.iloc[-NDAY - 1:-1])  # N日最低价的最低价
    range = max(HH - LC, HC - LL)
    buy_line = current_open + range * K1  # 上轨
    sell_line = current_open - range * K2  # 下轨
    print("当前开盘价: %f, 上轨: %f, 下轨: %f" % (current_open, buy_line, sell_line))
    return buy_line, sell_line


buy_line, sell_line = dual_thrust(quote, klines)  # 获取上下轨

while True:
    api.wait_update()
    if api.is_changing(klines.iloc[-1], ["datetime", "open"]):  # 新产生一根日线或开盘价发生变化: 重新计算上下轨
        buy_line, sell_line = dual_thrust(quote, klines)

    if api.is_changing(quote, "last_price"):
        if quote.last_price > buy_line:  # 高于上轨
            print("高于上轨,目标持仓 多头3手")
            target_pos.set_target_volume(3)  # 交易
        elif quote.last_price < sell_line:  # 低于下轨
            print("低于下轨,目标持仓 空头3手")
            target_pos.set_target_volume(-3)  # 交易
        else:
            print('未穿越上下轨,不调整持仓')
